import base64
import io
import math
import os
import traceback as tb
import zipfile
from typing import Literal, Optional

import click
from anvil.api.actors.materials import MaterialsObject
from anvil.api.core.sounds import BlocksJSONObject, MusicDefinition, SoundDefinition, SoundEvent
from anvil.api.core.textures import FlipBookTexturesObject, ItemTexturesObject, TerrainTexturesObject
from anvil.lib.blockbench import _Blockbench
from anvil.lib.config import CONFIG, ConfigPackageTarget, _AnvilConfig
from anvil.lib.format_versions import MANIFEST_VERSION, MODULE_MINECRAFT_SERVER, MODULE_MINECRAFT_SERVER_UI
from anvil.lib.lib import (
    CreateDirectory,
    File,
    FileExists,
    RemoveDirectory,
    process_subcommand,
    validate_namespace_project_name,
    zipit,
)
from anvil.lib.reports import ReportType
from anvil.lib.schemas import AddonObject, JsonSchemes
from anvil.lib.translator import AnvilTranslator
from halo import Halo
from PIL import Image

from ...__version__ import __version__


def resize(
    image: Image.Image,
    name: str,
    output: str,
    size,
    quality=95,
    dpi=72,
    overlay: Image.Image = None,
):
    resized = image.resize(size)

    if overlay:
        overlay = overlay.resize(size)
        resized.paste(overlay, mask=overlay.split()[3])

    resized.convert("RGB").save(
        os.path.join(output, name),
        dpi=(dpi, dpi),
        quality=quality,
    )


def extract_world_pack(extract_world: str | None = None):
    if extract_world != None and type(extract_world) is str:
        RemoveDirectory(CONFIG._WORLD_PATH)
        with zipfile.ZipFile(os.path.join("world", f"{extract_world}.mcworld"), "r") as zip_ref:
            zip_ref.extractall(CONFIG._WORLD_PATH)


def manifests(extract_world: str | None = None):
    release_list = [int(i) for i in CONFIG._RELEASE.split(".")]
    if extract_world != None and type(extract_world) is str:
        File(
            "manifest.json",
            JsonSchemes.manifest_world(release_list),
            CONFIG._WORLD_PATH,
            "w",
        )
        File(
            "world_resource_packs.json",
            JsonSchemes.world_packs(release_list, CONFIG._RP_UUID),
            CONFIG._WORLD_PATH,
            "w",
        )
        File(
            "world_behavior_packs.json",
            JsonSchemes.world_packs(release_list, CONFIG._BP_UUID),
            CONFIG._WORLD_PATH,
            "w",
        )


def scriptapi():
    if any([CONFIG._SCRIPT_API, CONFIG._SCRIPT_UI]):
        if not FileExists("tsconfig.json"):
            File(
                "tsconfig.json",
                JsonSchemes.tsconfig(CONFIG.BP_PATH),
                "",
                "w",
                False,
            )
        if not FileExists("esbuild.js") and "esbuild" in CONFIG._SCRIPT_BUNDLE_SCRIPT:
            File(
                "esbuild.js",
                JsonSchemes.esbuild_config_js(CONFIG.BP_PATH, CONFIG._MINIFY),
                "",
                "w",
                False,
            )
        if not FileExists("package.json"):
            File(
                "package.json",
                JsonSchemes.package_json(
                    CONFIG.PROJECT_NAME,
                    CONFIG._RELEASE,
                    CONFIG.PROJECT_DESCRIPTION,
                    CONFIG.COMPANY,
                ),
                "",
                "w",
                True,
            )


def addon_object_exception(object: AddonObject, e: Exception):
    full_traceback = tb.format_exc()
    creation_info = f"<{object.__class__.__name__} created from:\n{''.join(getattr(object, "_created_from", None))}>"
    click.echo(click.style(f"\r{'='*60}", fg="red"), err=True)
    click.echo(click.style(f"\r\nCreation Context:", fg="cyan"), err=True)
    click.echo(click.style(f"\r{creation_info}", fg="white"), err=True)
    click.echo(click.style(f"\r\nFull Traceback:", fg="cyan"), err=True)
    click.echo(click.style(f"\r{full_traceback}", fg="white"), err=True)
    click.echo(click.style(f"\r{'='*60}", fg="red"), err=True)
    click.echo(
        click.style(
            f"\rERROR EXPORTING OBJECT: {getattr(object, '_name', 'Unknown')}",
            fg="red",
        ),
        err=True,
    )
    click.echo(click.style(f"\r{'='*60}", fg="red"), err=True)
    click.echo(
        click.style(
            f"\rObject Type: {object.__class__.__name__}",
            fg="yellow",
        ),
        err=True,
    )
    click.echo(
        click.style(
            f"\rObject Name: {getattr(object, '_name', 'Unknown')}",
            fg="yellow",
        ),
        err=True,
    )
    click.echo(click.style(f"\rObject Content: {object._content}", fg="yellow"), err=True)
    click.echo(click.style(f"\rError Message: {str(e)}", fg="red"), err=True)
    click.echo(click.style(f"\r{'='*60}", fg="red"), err=True)


@Halo(text="Processing Art", spinner="dots")
def process_art(
    config: _AnvilConfig,
    apply_overlay: bool = False,
    zip: bool = True,
):
    """Process marketing art for packaging."""
    pack_icon_size = (256, 256)
    store_screenshot_size = (800, 450)
    marketing_screenshot_size = (1920, 1080)

    source = os.path.join("marketing")
    output_store = os.path.join("output", "Store Art")
    output_marketing = os.path.join("output", "Marketing Art")

    if zip:
        CreateDirectory(output_store)
        CreateDirectory(output_marketing)

        if FileExists(os.path.join(source, "keyart.png")):
            original = Image.open(os.path.join(source, "keyart.png")).convert("RGB")
            overlay = None

            if apply_overlay:
                overlay = Image.open(os.path.join(source, "keyart_overlay.png"))

            if config._TARGET == ConfigPackageTarget.WORLD:
                resize(
                    original,
                    "world_icon.jpeg",
                    config._WORLD_PATH,
                    store_screenshot_size,
                    95,
                    72,
                    overlay,
                )
            resize(
                original,
                f"{config.PROJECT_NAME}_Thumbnail_0.jpg",
                output_store,
                store_screenshot_size,
                95,
                72,
                overlay,
            )
            resize(
                original,
                f"{config.PROJECT_NAME}_MarketingKeyArt.jpg",
                output_marketing,
                marketing_screenshot_size,
                100,
                300,
                overlay,
            )

        else:
            raise FileNotFoundError("keyart.png not found in marketing directory. Please ensure the file exists.")

        if FileExists(os.path.join(source, "panorama.png")):
            original = Image.open(os.path.join(source, "panorama.png"))
            scale_factor = 450 / original.size[1]
            if (round(original.size[0] * scale_factor)) > 4000:
                raise RuntimeError(
                    f"panorama.png width exceeds 4000 pixels when resized to 450 pixels in height. Please use a smaller image. The recommended original size for this image would be {math.floor(4000 / scale_factor)}x450 pixels."
                )

            resize(
                original,
                f"{config.PROJECT_NAME}_panorama_0.jpg",
                output_store,
                (round(original.size[0] * scale_factor), 450),
                95,
                72,
            )

        else:
            click.echo(
                click.style(
                    f"\r[INFO]: panorama.png does not exist. It is optional but might have been left out unintentionally.",
                    fg="yellow",
                )
            )

        for i in range(999):
            if not FileExists(os.path.join(source, f"{i}.png")):
                if i < 5:
                    raise FileNotFoundError(f"{i}.png not found in marketing directory. Please ensure the file exists.")
                break
            else:
                original = Image.open(os.path.join(source, f"{i}.png"))
                resize(
                    original,
                    f"{config.PROJECT_NAME}_screenshot_{i}.jpg",
                    output_store,
                    store_screenshot_size,
                    95,
                    72,
                )
                resize(
                    original,
                    f"{config.PROJECT_NAME}_MarketingScreenshot_{i}.jpg",
                    output_marketing,
                    marketing_screenshot_size,
                    100,
                    300,
                )

        if FileExists(os.path.join(source, "partner_art.png")):
            original = Image.open(os.path.join(source, "partner_art.png"))

            resize(
                original,
                f"{config.PROJECT_NAME}_PartnerArt.jpg",
                output_marketing,
                marketing_screenshot_size,
                100,
                300,
            )

        else:
            raise FileNotFoundError("partner_art.png not found in marketing directory. Please ensure the file exists.")

        if FileExists(os.path.join(source, "pack_icon.png")):
            original = Image.open(os.path.join(source, "pack_icon.png"))
            resize(
                original,
                f"{config.PROJECT_NAME}_packicon_0.jpg",
                output_store,
                pack_icon_size,
            )
        else:
            raise FileNotFoundError("pack_icon.png not found in marketing directory. Please ensure the file exists.")


@Halo(text="Generating technical notes", spinner="dots")
def generate_technical_notes_pdf(config: _AnvilConfig):
    """Generates a technical notes PDF that contains information about included entities, blocks, items, sounds and more."""
    from reportlab.lib import colors
    from reportlab.lib.pagesizes import A4
    from reportlab.lib.styles import getSampleStyleSheet
    from reportlab.lib.units import cm
    from reportlab.platypus import Paragraph, SimpleDocTemplate, Spacer, Table, TableStyle

    def add_table(section_name: str, data: dict[bool, set[str]]):
        title_style.spaceBefore = 0
        title_style.fontSize = 14
        title_style.textColor = colors.royalblue
        title = Paragraph(section_name, title_style)

        converted_data = []
        vanilla_true_rows = []

        for idx, (row, columns) in enumerate(data.items()):
            vals = []
            for col_name, col_values in columns.items():
                value_string = "<br/>".join(col_values)
                if col_name != "vanilla":
                    vals.append(Paragraph(value_string, styles["Normal"]))
                elif value_string == "True":
                    vanilla_true_rows.append(idx)
            converted_data.append([Paragraph(row, styles["Normal"]), *vals])

        table = Table(
            converted_data,
            hAlign="LEFT",
            colWidths=doc.width / len(converted_data[0]),
        )

        style_commands = [
            ("BACKGROUND", (0, 0), (-1, 0), colors.lightblue),
            ("TEXTCOLOR", (0, 0), (-1, 0), colors.black),
            ("ALIGN", (0, 0), (-1, -1), "LEFT"),
            ("VALIGN", (0, 0), (-1, -1), "TOP"),
            ("FONTSIZE", (0, 0), (-1, 0), 12),
            ("FONTNAME", (0, 0), (-1, 0), "Helvetica"),
            ("BOTTOMPADDING", (0, 0), (-1, 0), 12),
            ("BACKGROUND", (0, 1), (-1, -1), colors.white),
            ("GRID", (0, 0), (-1, -1), 1, colors.black),
        ]
        for row in vanilla_true_rows:
            style_commands.append(("BACKGROUND", (0, row), (-1, row), colors.lightgreen))
        table.setStyle(TableStyle(style_commands))

        return title, Spacer(1, 0.3 * cm), table, Spacer(1, 1 * cm)

    doc = SimpleDocTemplate(
        os.path.join("output", "technical_notes.pdf"),
        pagesize=A4,
        leftMargin=1 * cm,
        rightMargin=1 * cm,
        topMargin=1 * cm,
        bottomMargin=1 * cm,
        title=f"{config.DISPLAY_NAME} Technical Notes",
        author=config.COMPANY,
        subject=f"{config.DISPLAY_NAME} Technical Notes",
        creator=f"Anvil@stark_lg {__version__}",
    )
    styles = getSampleStyleSheet()
    title_style = styles["Heading1"]
    body_style = styles["BodyText"]
    bullet_style = styles["Bullet"]

    elements = [
        Paragraph(f"{config.DISPLAY_NAME}:", title_style),
        Paragraph(f"Developed by: {config.COMPANY}", body_style),
        Paragraph(
            f'Generated with <a href="https://github.com/StarkTMA/Anvil"><u><font color="blue">StarkTMA/Anvil {__version__}</font></u></a>',
            body_style,
        ),
        Spacer(1, 1 * cm),
        Paragraph("General information:", title_style),
        Paragraph(
            "The following technical notes have been entirely generated from source code using Anvil.",
            body_style,
        ),
        Paragraph(
            "Features overwriting vanilla defaults will be highlighted in green.",
            bullet_style,
            "*",
        ),
        Spacer(1, 1 * cm),
        *add_table("Entities:", config.Report.dict[ReportType.ENTITY]),
        *add_table("Attachables:", config.Report.dict[ReportType.ATTACHABLE]),
        *add_table("Items:", config.Report.dict[ReportType.ITEM]),
        *add_table("Blocks:", config.Report.dict[ReportType.BLOCK]),
        *add_table("Particles:", config.Report.dict[ReportType.PARTICLE]),
        *add_table("Sounds:", config.Report.dict[ReportType.SOUND]),
    ]

    doc.build(elements)


@Halo(text="Packaging ZIP", spinner="dots")
def package_zip_core(
    config: _AnvilConfig,
    apply_overlay: bool = False,
    generate_technical_notes: bool = False,
):
    """Core logic for packaging the project into a zip file for Marketplace."""
    process_art(config, apply_overlay)

    content_structure = {
        os.path.join("output", "Store Art"): os.path.join("Store Art"),
        os.path.join("output", "Marketing Art"): os.path.join("Marketing Art"),
    }

    if config._TARGET == ConfigPackageTarget.ADDON:
        content_structure.update(
            {
                config.RP_PATH: os.path.join(
                    "Content",
                    "resource_packs",
                    f"RP_{config._PASCAL_PROJECT_NAME}",
                ),
                config.BP_PATH: os.path.join(
                    "Content",
                    "behavior_packs",
                    f"BP_{config._PASCAL_PROJECT_NAME}",
                ),
            }
        )

    else:
        content_structure.update(
            {
                config.RP_PATH: os.path.join(
                    "Content",
                    "world_template",
                    "resource_packs",
                    f"RP_{config._PASCAL_PROJECT_NAME}",
                ),
                config.BP_PATH: os.path.join(
                    "Content",
                    "world_template",
                    "behavior_packs",
                    f"BP_{config._PASCAL_PROJECT_NAME}",
                ),
                os.path.join(config._WORLD_PATH, "texts"): os.path.join("Content", "world_template", "texts"),
                os.path.join(config._WORLD_PATH, "level.dat"): os.path.join("Content", "world_template"),
                os.path.join(config._WORLD_PATH, "levelname.txt"): os.path.join("Content", "world_template"),
                os.path.join(config._WORLD_PATH, "manifest.json"): os.path.join("Content", "world_template"),
                os.path.join(config._WORLD_PATH, "world_icon.jpeg"): os.path.join("Content", "world_template"),
                os.path.join(config._WORLD_PATH, "world_behavior_packs.json"): os.path.join(
                    "Content", "world_template"
                ),
                os.path.join(config._WORLD_PATH, "world_resource_packs.json"): os.path.join(
                    "Content", "world_template"
                ),
            }
        )
        if not config._RANDOM_SEED:
            content_structure.update(
                {
                    os.path.join(config._WORLD_PATH, "db"): os.path.join("Content", "world_template", "db"),
                }
            )

    zipit(
        os.path.join("output", f"{config.PROJECT_NAME}.zip"),
        content_structure,
    )

    RemoveDirectory(os.path.join("output", "Store Art"))
    RemoveDirectory(os.path.join("output", "Marketing Art"))

    if generate_technical_notes:
        generate_technical_notes_pdf(config)


@Halo(text="Packaging mcaddon", spinner="dots")
def mcaddon_core(config):
    """Core logic for packaging the project into a .mcaddon file."""
    process_art(config, False, False)

    content_structure = {
        config.RP_PATH: f"RP_{config.PROJECT_NAME}",
        config.BP_PATH: f"BP_{config.PROJECT_NAME}",
    }

    zipit(
        os.path.join("output", f"{config.PROJECT_NAME}.mcaddon"),
        content_structure,
    )


@Halo(text="Packaging mcworld", spinner="dots")
def mcworld_core(config: _AnvilConfig):
    """Core logic for packaging the project into a .mcworld file."""
    process_art(config, False, False)

    content_structure = {
        config.RP_PATH: os.path.join("resource_packs", f"RP_{config.PROJECT_NAME}"),
        config.BP_PATH: os.path.join("behavior_packs", f"BP_{config.PROJECT_NAME}"),
        os.path.join(config._WORLD_PATH, "texts"): "texts",
        os.path.join(config._WORLD_PATH, "level.dat"): "",
        os.path.join(config._WORLD_PATH, "levelname.txt"): "",
        os.path.join(config._WORLD_PATH, "manifest.json"): "",
        os.path.join(config._WORLD_PATH, "world_icon.jpeg"): "",
        os.path.join(config._WORLD_PATH, "world_behavior_packs.json"): "",
        os.path.join(config._WORLD_PATH, "world_resource_packs.json"): "",
    }

    zipit(
        os.path.join("output", f"{config.PROJECT_NAME}.mcworld"),
        content_structure,
    )


@Halo(text="Compiling", spinner="dots")
def compile_objects(anvil: "_Anvil", extract_world: str = None):
    extract_world_pack(extract_world)
    manifests(extract_world)
    scriptapi()

    ItemTexturesObject().queue()
    TerrainTexturesObject().queue()
    FlipBookTexturesObject().queue()
    SoundDefinition().queue()
    SoundEvent().queue()
    MusicDefinition().queue()
    BlocksJSONObject().queue()
    MaterialsObject().queue()
    _Blockbench._export()
    anvil._queue(AnvilTranslator())
    ManifestBP().queue()
    ManifestRP().queue()

    from anvil.api.blocks.blocks import _PermutationComponents

    if _PermutationComponents._count > 10000:
        if CONFIG._TARGET == ConfigPackageTarget.ADDON:
            raise RuntimeError(
                f"\rTotal Block permutations exceeded 10000 ({_PermutationComponents._count}). Addons must not exceed this limit."
            )
        else:
            click.echo(
                click.style(
                    f"\r[INFO]: Total Block permutations exceeded 10000 ({_PermutationComponents._count}). This may cause issues when submitting to the Marketplace.",
                    fg="yellow",
                )
            )

    for object in anvil._objects_list:
        try:
            object._export()
        except Exception as e:
            addon_object_exception(object, e)
            return

    if CONFIG._TARGET == ConfigPackageTarget.ADDON:
        if len(CONFIG._RP_UUID) > 1:
            raise RuntimeError(
                "Multiple resource pack UUIDs found. Please ensure only one UUID is set for the resource pack."
            )
        if len(CONFIG._BP_UUID) > 1:
            raise RuntimeError(
                "Multiple behavior pack UUIDs found. Please ensure only one UUID is set for the behavior pack."
            )

    if CONFIG._SCRIPT_API:
        process_subcommand(
            CONFIG._SCRIPT_BUNDLE_SCRIPT + f' --outdir="{os.path.join(CONFIG.BP_PATH, "scripts")}"',
            "Building scripts error",
        )

    source = os.path.join("marketing")
    pack_icon_size = (256, 256)

    if FileExists(os.path.join(source, "pack_icon.png")):
        original = Image.open(os.path.join(source, "pack_icon.png"))
        resize(original, "pack_icon.png", CONFIG.BP_PATH, pack_icon_size)
        resize(original, "pack_icon.png", CONFIG.RP_PATH, pack_icon_size)

    else:
        click.echo(
            click.style(
                f"\r[INFO]: pack_icon.png not found in marketing directory. Using placeholder pack icon.",
                fg="yellow",
            )
        )
        placeholder_base64 = " "
        placeholder_bytes = base64.b64decode(placeholder_base64)
        placeholder_image = Image.open(io.BytesIO(placeholder_bytes))

        resize(placeholder_image, "pack_icon.png", CONFIG.BP_PATH, pack_icon_size)
        resize(placeholder_image, "pack_icon.png", CONFIG.RP_PATH, pack_icon_size)

    anvil._compiled = True


@Halo(text="Translating", spinner="dots")
def translate(languages: Optional[list[str]] = None) -> None:
    """Translates the project."""
    AnvilTranslator().auto_translate_all(languages)


class ManifestRP(AddonObject):
    _instance = None
    _extension = ".json"
    _path = os.path.join(CONFIG.RP_PATH)
    _setting_ids: set[str] = set()

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(ManifestRP, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        """Initializes a ManifestRP instance."""
        if not hasattr(self, "_initialized"):
            super().__init__("manifest")
            self.content(JsonSchemes.manifest_rp(version=CONFIG._RELEASE))
            self._initialized = True

    @property
    def settings(self) -> "_ManifestSettings":
        """Returns a Manifest Settings object to add settings to the manifest."""
        return _ManifestSettings(self)

    def queue(self):
        if CONFIG._PBR:
            self._content.update({"capabilities": ["pbr"]})

        if CONFIG._TARGET == "addon":
            self._content["header"]["pack_scope"] = "world"
            self._content["metadata"]["product_type"] = "addon"

        return super().queue()


class ManifestBP(AddonObject):
    _instance = None
    _extension = ".json"
    _path = os.path.join(CONFIG.BP_PATH)
    _setting_ids: set[str] = set()

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(ManifestBP, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        """Initializes a ManifestBP instance."""
        if not hasattr(self, "_initialized"):
            super().__init__("manifest")
            self.content(JsonSchemes.manifest_bp(version=CONFIG._RELEASE))
            self._initialized = True

    @property
    def settings(self) -> "_ManifestSettings":
        """Returns a Manifest Settings object to add settings to the manifest."""
        if not CONFIG._PREVIEW:
            raise RuntimeError("Behaviour Pack Settings are currently only supported in current preview builds.")
        return _ManifestSettings(self)

    def queue(self):
        if CONFIG._SCRIPT_API:
            self._content["modules"].append(
                {
                    "uuid": CONFIG._SCRIPT_MODULE_UUID,
                    "version": CONFIG._RELEASE,
                    "type": "script",
                    "language": "javascript",
                    "entry": "scripts/main.js",
                }
            )
            self._content["dependencies"].append(
                {
                    "module_name": "@minecraft/server",
                    "version": MODULE_MINECRAFT_SERVER,
                }
            )
            if CONFIG._SCRIPT_UI:
                self._content["dependencies"].append(
                    {
                        "module_name": "@minecraft/server-ui",
                        "version": MODULE_MINECRAFT_SERVER_UI,
                    }
                )

        if CONFIG._TARGET == "addon":
            self._content["header"]["pack_scope"] = "world"
            self._content["metadata"]["product_type"] = "addon"

        return super().queue()


class _ManifestSettings:
    def __init__(self, manifest: ManifestRP | ManifestBP):
        if MANIFEST_VERSION < 3:
            raise RuntimeError("Settings labels are only supported in manifest version 3 and above.")
        self.manifest = manifest

        # Initialize settings in _content if not present
        if "settings" not in self.manifest._content:
            self.manifest._content["settings"] = []

    def label(self, label: str):
        self.manifest._content["settings"].append({"type": "label", "text": label})
        return self

    def toggle(self, id: str, text: str, default: bool = True):
        if id in self.manifest._setting_ids:
            raise ValueError(f"Setting ID '{id}' is already used.")

        self.manifest._setting_ids.add(id)
        self.manifest._content["settings"].append(
            {
                "type": "toggle",
                "name": f"{CONFIG.NAMESPACE}:{id}",
                "text": text,
                "default": default,
            }
        )
        return self

    def slider(self, id: str, text: str, min: int, max: int, step: int, default: int):
        if min >= max:
            raise ValueError(f"Min value '{min}' must be less than max '{max}'.")

        if step <= 0:
            raise ValueError(f"Step value '{step}' must be greater than 0.")

        if not (min <= default <= max):
            raise ValueError(f"Default value '{default}' must be between min '{min}' and max '{max}'.")

        if id in self.manifest._setting_ids:
            raise ValueError(f"Setting ID '{id}' is already used.")

        self.manifest._setting_ids.add(id)
        self.manifest._content["settings"].append(
            {
                "type": "slider",
                "name": f"{CONFIG.NAMESPACE}:{id}",
                "text": text,
                "min": min,
                "max": max,
                "step": step,
                "default": default,
            }
        )
        return self

    def dropdown(self, id: str, text: str, options: list[dict[str, str]], default: str):
        if not isinstance(options, list) or len(options) == 0:
            raise ValueError("Options must be a non-empty list of dictionaries.")

        if default not in [key for option in options for key in option.keys()]:
            raise ValueError(f"Default value '{default}' is not in the options list.")

        if id in self.manifest._setting_ids:
            raise ValueError(f"Setting ID '{id}' is already used.")

        self.manifest._setting_ids.add(id)
        self.manifest._content["settings"].append(
            {
                "type": "dropdown",
                "name": f"{CONFIG.NAMESPACE}:{id}",
                "text": text,
                "options": [{"name": k, "text": v} for option in options for k, v in option.items()],
                "default": default,
            }
        )
        return self


class _Anvil:
    """A class representing an Anvil instance."""

    _instance = None
    _objects_list: list[AddonObject] = []

    def _queue(self, object: AddonObject):
        """Queues an object to be compiled."""
        self._objects_list.append(object)

    def __init__(self):
        """Initializes an Anvil instance.

        Parameters:
            config (_Config): The config of the Anvil instance.
        """
        if _Anvil._instance is None:
            _Anvil._instance = self
        else:
            raise Exception("Anvil instance already exists.")

        self._compiled = False
        self.config = CONFIG
        validate_namespace_project_name(CONFIG.NAMESPACE, CONFIG.PROJECT_NAME)

    def translate(self, languages: Optional[list[str]] = None) -> None:
        """Translates the project."""
        translate(languages)

    def compile(self, extract_world: str = None) -> None:
        """Compiles the project."""
        compile_objects(self, extract_world)

    def package_zip(
        self,
        apply_overlay: bool = False,
        generate_technical_notes: bool = False,
    ) -> None:
        """Packages the project into a zip file for Marketplace.

        Parameters:
            apply_overlay (bool, optional): Whether to apply the overlay to the marketing art. Defaults to False.
            generate_technical_notes (bool, optional): Whether to generate technical notes PDF. Defaults to False.
        """
        if not self._compiled:
            raise RuntimeError("Project must be compiled before packaging.")

        package_zip_core(self.config, apply_overlay, generate_technical_notes)

    def mcaddon(self):
        """Packages the project into a .mcaddon file."""
        if not self._compiled:
            raise RuntimeError("Project must be compiled before packaging.")

        mcaddon_core(self.config)

    def mcworld(self):
        """Packages the project into a .mcworld file."""
        if not self._compiled:
            raise RuntimeError("Project must be compiled before packaging.")

        mcworld_core(self.config)


ANVIL = _Anvil()



from typing import Literal, TypedDict

type Vector2D = tuple[float, float]
type Vector3D = tuple[float, float, float]
type Vector4D = tuple[float, float, float, float]
type Vector2DInt = tuple[int, int]
type Vector3DInt = tuple[int, int, int]
type Vector4DInt = tuple[int, int, int, int]

type RGB = Vector3D
type RGBA = Vector4D
type RGB255 = Vector3DInt
type RGBA255 = Vector4DInt
type HexRGB = str
type HexRGBA = str
type Color = RGB | RGBA | RGB255 | RGBA255 | HexRGB | HexRGBA

type Coordinate = Vector3D
type Coordinates = tuple[Vector3D, Vector3D, Vector3D]
type RelativePosition = tuple[int | Literal["^", "~"], int | Literal["^", "~"], int | Literal["^", "~"]]
type RelativeRotation = tuple[int | Literal["~"], int | Literal["~"]]

type Component = str
type Event = str
type Identifier = str
type Level = tuple[float, float]
type Range = tuple[float, float]
type Rotation = tuple[Coordinate, Coordinate]
type Seconds = int
type Tick = int
type Target = Literal["@p", "@r", "@a", "@e", "@s", "@c", "@v", "@initiator"]
type StructureProcessors = Literal[
    "minecraft:block_ignore",
    "minecraft:protected_blocks",
    "minecraft:capped",
    "minecraft:rule",
]
type Block = str


class ConstantIntProvider(TypedDict):
    value: int


class UniformIntProvider(TypedDict):
    min: int
    max: int



import json
import os

from anvil.api.pbr.pbr import TextureComponents, TextureSet
from anvil.lib.config import CONFIG
from anvil.lib.lib import Color, CopyFiles, FileExists
from anvil.lib.reports import ReportType
from anvil.lib.schemas import AddonObject


class Particle(AddonObject):
    _extension = ".particle.json"
    _path = os.path.join(CONFIG.RP_PATH, "particles")
    _object_type = "Particle"

    def __init__(
        self,
        particle: str,
        component: TextureComponents,
    ):
        """Create a particle with optional PBR texture set support.

        Args:
            particle: Name of the particle
            color_texture: color texture name
            normal_texture: normal texture name (optional)
            height_texture: height texture name (optional, mutually exclusive with normal)
            mer_texture: MER texture name (optional)
            mers_texture: MER+subsurface texture name (optional)
        """
        super().__init__(particle)
        self._texture_set: TextureSet = None
        self._color_texture = component.color

        if not FileExists(os.path.join("assets", "particles", f"{self._name}.particle.json")):
            raise FileNotFoundError(
                f"Particle file '{self._name}.particle.json' does not exist in 'assets/particles'. {self._object_type}[{self._name}]"
            )

        with open(os.path.join("assets", "particles", f"{self._name}.particle.json"), "r") as file:
            self._content = json.loads(file.read())
            if self._content["particle_effect"]["description"]["identifier"] != self._name:
                raise ValueError(
                    f"Particle identifier mismatch: expected '{self._name}', got '{self._content['particle_effect']['description']['identifier']}'"
                )

        self._texture_set = TextureSet(self._color_texture, "particle")
        self._texture_set.set_particle_textures(component)

        self._content["particle_effect"]["description"]["basic_render_parameters"]["texture"] = os.path.join(
            "textures",
            CONFIG.NAMESPACE,
            CONFIG.PROJECT_NAME,
            "particle",
            self._color_texture,
        )

        self._content["particle_effect"]["description"]["identifier"] = f"{CONFIG.NAMESPACE}:{self._name}"

    def queue(self):
        CONFIG.Report.add_report(
            ReportType.PARTICLE,
            vanilla=False,
            col0=self._name.replace("_", " ").title(),
            col1=f"{CONFIG.NAMESPACE}:{self._name}",
        )

        return super().queue()

    def _export(self):
        # Export texture set if configured
        if self._texture_set is not None:
            self._texture_set.queue()

        super()._export()



import os
from pathlib import Path
from typing import Dict, List, Optional

import click
import pandas as pd
from deep_translator import GoogleTranslator

from anvil.lib.config import CONFIG
from anvil.lib.lib import File
from anvil.lib.schemas import JsonSchemes


class AnvilTranslator:
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(AnvilTranslator, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        """
        Initialize the translator with an Excel file path.
        """
        if hasattr(self, "_initialized") and self._initialized:
            return
        self._excel_file_path = Path("localization.xlsx")
        self._source_language = "en_US"
        self._translated = False
        self.translated_languages = [self._source_language]
        self._initialized = True
        self.config = CONFIG

        # Runtime storage for user-registered entries
        self._runtime_entries: Dict[str, str] = {}
        
        # Original state from file
        self._original_entries: Dict[str, str] = {}
        
        if self._excel_file_path.exists():
            try:
                # Load original en_US entries to track deletions/modifications
                df = pd.read_excel(self._excel_file_path, sheet_name=self._source_language)
                self._original_entries = dict(zip(df["Key"], df["Value"]))
            except ValueError:
                pass
        else:
            with pd.ExcelWriter(self._excel_file_path, engine="openpyxl") as writer:
                for lang_code in JsonSchemes.languages():
                    df = pd.DataFrame(columns=["Key", "Value"])
                    df.to_excel(writer, sheet_name=lang_code, index=False)

        self.add_localization_entry("pack.name", self.config.DISPLAY_NAME)
        self.add_localization_entry(
            "pack.project_description", self.config.PROJECT_DESCRIPTION
        )
        self.add_localization_entry(
            "pack.resource_description", self.config.RESOURCE_DESCRIPTION
        )
        self.add_localization_entry(
            "pack.behaviour_description", self.config.BEHAVIOUR_DESCRIPTION
        )

    def _get_languages(self) -> List[str]:
        """Returns a list of all languages."""
        languages = [lang for lang in JsonSchemes.languages()]
        return languages

    def _get_localization_entries(self, language: str) -> Dict[str, str]:
        """
        Returns a dictionary of localization entries for a specific language.
        """
        try:
            df = pd.read_excel(self._excel_file_path, sheet_name=language)
            return dict(zip(df["Key"], df["Value"]))
        except ValueError:
            return {}

    def _sync_excel(self) -> None:
        """
        Synchronizes the runtime entries with the Excel file.
        """
        if not self._runtime_entries:
            return

        # Load all existing sheets
        all_dfs = {}
        for language in self._get_languages():
            try:
                df = pd.read_excel(self._excel_file_path, sheet_name=language)
                all_dfs[language] = df
            except ValueError:
                all_dfs[language] = pd.DataFrame(columns=["Key", "Value"])

        original_set = set(self._original_entries.keys())
        runtime_set = set(self._runtime_entries.keys())
        
        # Identify changes
        new_keys = runtime_set - original_set
        deleted_keys = original_set - runtime_set
        shared_keys = runtime_set.intersection(original_set)

        # 1. Handle New Keys
        if new_keys:
            new_rows_en = pd.DataFrame({"Key": list(new_keys), "Value": [self._runtime_entries[k] for k in new_keys]})
            all_dfs[self._source_language] = pd.concat([all_dfs[self._source_language], new_rows_en], ignore_index=True)
            
            for lang, df in all_dfs.items():
                if lang == self._source_language: 
                    continue
                new_rows_other = pd.DataFrame({"Key": list(new_keys), "Value": [""] * len(new_keys)})
                all_dfs[lang] = pd.concat([df, new_rows_other], ignore_index=True)

        # 2. Handle Deleted Keys
        if deleted_keys:
            for lang in all_dfs:
                all_dfs[lang] = all_dfs[lang][~all_dfs[lang]["Key"].isin(deleted_keys)]

        # 3. Handle Modified Values
        for key in shared_keys:
            original_val = self._original_entries[key]
            runtime_val = self._runtime_entries[key]
            
            if original_val != runtime_val:
                # Update en_US
                df_en = all_dfs[self._source_language]
                df_en.loc[df_en["Key"] == key, "Value"] = runtime_val
                
                # Clear other languages
                for lang in all_dfs:
                    if lang == self._source_language:
                        continue
                    df_lang = all_dfs[lang]
                    if key in df_lang["Key"].values:
                         df_lang.loc[df_lang["Key"] == key, "Value"] = pd.NA

        # Write back to Excel
        with pd.ExcelWriter(
            self._excel_file_path,
            engine="openpyxl",
            mode="a",
            if_sheet_exists="replace",
        ) as writer:
            for lang, df in all_dfs.items():
                df.to_excel(writer, sheet_name=lang, index=False)
                worksheet = writer.sheets[lang]
                if not df.empty:
                    key_lengths = [len(str(k)) for k in df["Key"] if pd.notna(k)]
                    max_key_length = max(key_lengths) if key_lengths else 10
                    worksheet.column_dimensions["A"].width = max(
                        max_key_length + 5, 20
                    )
                worksheet.column_dimensions["B"].width = 20
        
        # Update original entries after sync
        self._original_entries = self._runtime_entries.copy()

    def add_localization_entry(self, key: str, value: str) -> None:
        """
        Add a new localization entry to the runtime memory.

        Parameters:
            key (str): The localization key
            value (str): The English value
        """
        self._runtime_entries[key] = value

    def add_entries(self, entries: Dict[str, str]) -> None:
        """
        Add multiple localization entries to the en_US sheet.

        Parameters:
            entries (Dict[str, str]): A dictionary of localization entries
        """
        for key, value in entries.items():
            self.add_localization_entry(key, value)

    def get_localization_value(self, key: str) -> Optional[str]:
        """
        Get the localization value for a specific key from the source language (en_US).
        """
        try:
            df = pd.read_excel(self._excel_file_path, sheet_name=self._source_language)
            return df.loc[df["Key"] == key, "Value"].iloc[0]
        except (ValueError, IndexError):
            return None

    def auto_translate_all(self, languages: Optional[list[str]] = None) -> None:
        """
        Automatically translate all entries from source language to target languages.
        """
        self._sync_excel()

        def parse(lang: str):
            lang = lang.replace("zh_CN", "zh-CN")
            lang = lang.replace("zh_TW", "zh-TW")
            lang = lang.replace("nb_NO", "no").split("_")[0]
            lang = lang.replace("en_US", "en")
            lang = lang.replace("en_GB", "en")
            return lang

        if languages is None:
            languages = self._get_languages()

        source_entries = self._get_localization_entries(self._source_language)

        with pd.ExcelWriter(
            self._excel_file_path,
            engine="openpyxl",
            mode="a",
            if_sheet_exists="replace",
        ) as writer:
            for target_lang in languages:

                if target_lang == self._source_language:
                    continue

                existing_entries = self._get_localization_entries(target_lang)
                keys_to_translate = []
                values_to_translate = []

                for key, value in source_entries.items():
                    if (
                        key not in existing_entries
                        or existing_entries[key] == ""
                        or pd.isna(existing_entries[key])
                    ):
                        keys_to_translate.append(key)
                        values_to_translate.append(value)

                if not values_to_translate:
                    continue

                translator = GoogleTranslator(source="en", target=parse(target_lang))
                translated_values = []
                batch_size = 50
                for i in range(0, len(values_to_translate), batch_size):
                    batch = values_to_translate[i : i + batch_size]
                    try:
                        translated_values.extend(translator.translate_batch(batch))
                    except Exception as e:
                        click.echo(
                            click.style(
                                f"[INFO]: Translation error for {target_lang}: {e}",
                                fg="red",
                            )
                        )
                        translated_values.extend(batch)

                translation_map = dict(zip(keys_to_translate, translated_values))

                all_keys = list(source_entries.keys())
                all_values = []

                for key in all_keys:
                    if key in existing_entries and not pd.isna(existing_entries[key]):
                        all_values.append(existing_entries[key])
                    else:
                        all_values.append(translation_map[key])

                df = pd.DataFrame({"Key": all_keys, "Value": all_values})
                df.to_excel(writer, sheet_name=target_lang, index=False)

        self._cleanup_orphaned_keys()

        self._translated = True
        self.translated_languages.extend(languages)

    def _export(self) -> None:
        """
        Export translations to Anvil's .lang file format.
        """
        self._sync_excel()

        languages = []
        skipped = []

        for lang_code in self._get_languages():
            entries = self._get_localization_entries(lang_code)
            if not entries:
                continue

            if any(not value or pd.isna(value) for value in entries.values()):
                skipped.append(lang_code)
                continue
            languages.append(lang_code)

            lang_content = []
            ordered_keys = [
                "pack.name",
                "pack.project_description",
                "pack.resource_description",
                "pack.behaviour_description",
            ]
            for key in ordered_keys:
                if key in entries:
                    lang_content.append(f"{key}={entries[key]}")

            remaining_entries = {
                k: v for k, v in entries.items() if k not in ordered_keys
            }
            for key, value in sorted(remaining_entries.items()):
                lang_content.append(f"{key}={value}")

            bp_content = [
                lang_content[0],
                lang_content[3].replace("behaviour_description", "description"),
            ]
            world_content = [
                lang_content[0],
                lang_content[1].replace("project_description", "description"),
            ]

            rp_content = lang_content.copy()
            rp_content[2] = lang_content[2].replace(
                "resource_description", "description"
            )
            del rp_content[1]
            del rp_content[2]

            File(
                f"{lang_code}.lang",
                "\n".join(rp_content),
                os.path.join(self.config.RP_PATH, "texts"),
                "w",
            )
            File(
                f"{lang_code}.lang",
                "\n".join(bp_content),
                os.path.join(self.config.BP_PATH, "texts"),
                "w",
            )

            if self.config._TARGET == "world":
                File(
                    f"{lang_code}.lang",
                    "\n".join(world_content),
                    os.path.join(self.config._WORLD_PATH, "texts"),
                    "w",
                )

        if skipped:
            click.echo(
                click.style(
                    f"\r[INFO]: Skipping [{', '.join(skipped)}] - contains empty values",
                    fg="yellow",
                )
            )
        File(
            "languages.json", languages, os.path.join(self.config.BP_PATH, "texts"), "w"
        )
        File(
            "languages.json", languages, os.path.join(self.config.RP_PATH, "texts"), "w"
        )
        if self.config._TARGET == "world":
            File(
                "languages.json",
                languages,
                os.path.join(self.config._WORLD_PATH, "texts"),
                "w",
            )



import os
from typing import Any, Dict, List, Literal, Mapping

from anvil import ANVIL
from anvil.api.actors.actors import _Components
from anvil.api.blocks.components import BlockDisplayName, BlockGeometry, BlockMaterialInstance
from anvil.api.core.components import _BaseComponent
from anvil.api.core.core import SoundEvent
from anvil.api.core.enums import (
    BlockInteractiveSoundEvent,
    BlockSoundEvent,
    ConnectionTrait,
    ItemCategory,
    ItemGroups,
    PlacementDirectionTrait,
    PlacementPositionTrait,
)
from anvil.api.logic.molang import Molang
from anvil.api.vanilla.blocks import MinecraftBlockTags
from anvil.lib.config import CONFIG
from anvil.lib.reports import ReportType
from anvil.lib.schemas import AddonObject, JsonSchemes, MinecraftBlockDescriptor, MinecraftDescription
from anvil.lib.translator import AnvilTranslator

__all__ = ["Block"]


class _tag:
    """Represents a block tag component.

    Handles the creation and management of Minecraft block tags
    with associated dependencies and clashes.
    """

    def __init__(self, tag: MinecraftBlockTags):
        """Initializes a block tag component.

        Args:
            tag (MinecraftBlockTags): The vanilla tag to assign to the block.
        """
        self._identifier = f"tag:{tag}"
        self._dependencies: List["_BaseComponent"] = []
        self._clashes: List["_BaseComponent"] = []
        self._component: Dict[str, Any] = {f"tag:{tag}": {}}

    def __iter__(self):
        """Iterates over the tag component items.

        Returns:
            iterator: An iterator over the component dictionary items.
        """
        return iter(self._component.items())


# Core
class _PermutationComponents(_Components):
    _count = 0

    def __init__(self, condition: str | Molang = None):
        """The permutation components.

        Args:
            condition (str | Molang, optional): The condition for the permutation.
        """
        super().__init__()
        _PermutationComponents._count += 1
        self._component_group_name = "components"
        self._condition = condition

    def tag(self, *tags: MinecraftBlockTags):
        """The tags for the block.

        Args:
            *tags (MinecraftBlockTags): The tags for the block.

        ## Documentation reference:
            https://learn.microsoft.com/en-gb/minecraft/creator/reference/content/blockreference/examples/blocktags
        """
        for tag in tags:
            self._set(_tag(tag))

        return self

    def _export(self):
        cmp = super()._export()
        if self._condition:
            cmp["condition"] = self._condition
        return cmp


class _BlockTraits:
    def __init__(self) -> None:
        self._traits = {}

    def placement_direction(
        self,
        *,
        y_rotation_offset: float = 0,
        blocks_to_corner_with: list[MinecraftBlockDescriptor] = None,
        traits: list[PlacementDirectionTrait] = None,
    ):
        """can add states containing information about the player's rotation when the block is placed.

        Parameters:
            y_rotation_offset (float, optional): The y rotation offset. Defaults to 0.
            blocks_to_corner_with (list[MinecraftBlockDescriptor], optional): A list of blocks that when placed next to this block, will cause the block to rotate to face the corner between them. Defaults to None.
            traits (list[PlacementDirectionTrait], optional): The traits for the block. Defaults to None.

        ## Documentation reference:
            https://learn.microsoft.com/en-gb/minecraft/creator/reference/content/blockreference/examples/blocktraits#placement_direction-example
        """

        if blocks_to_corner_with and PlacementPositionTrait.CornerAndCardinal not in traits:
            raise ValueError(
                "blocks_to_corner_with can only be used if PlacementPositionTrait.CornerAndCardinal is in traits."
            )

        self._traits["minecraft:placement_direction"] = {
            "enabled_states": traits,
            "y_rotation_offset": y_rotation_offset,
            "blocks_to_corner_with": blocks_to_corner_with,
        }

    def placement_position(self, traits: list[PlacementPositionTrait]):
        """Can add states containing information about the position of the block when it is placed.

        Parameters:
            traits (list[PlacementPositionTrait]): The traits for the block.

        ## Documentation reference:
            https://learn.microsoft.com/en-gb/minecraft/creator/reference/content/blockreference/examples/blocktraits#placement_position-example
        """
        self._traits["minecraft:placement_position"] = {"enabled_states": traits}

    def connection(self, traits: list[ConnectionTrait]):
        """Can add states containing information about the connection of the block to other blocks.

        Parameters:
            traits (list[ConnectionTrait]): The traits for the block.

        ## Documentation reference:
            https://learn.microsoft.com/en-gb/minecraft/creator/reference/content/blockreference/examples/blocktraits#connection_trait_example
        """
        self._traits["minecraft:connection"] = {"enabled_states": traits}

    def multi_block(self, direction: Literal["up", "down"], part_count: int = 2):
        """Defines a block composed of multiple block parts.

        Multi blocks treat all parts as a single block, similar to doors. When the
        block uses ``minecraft:selection_box`` with a raw ``true`` value, Bedrock
        expands the selection outline by combining the AABBs of each part.

        Parameters:
            direction (Literal["up", "down"]): Direction the parts are placed from.
            part_count (int, optional): Number of parts in the multi block. Valid range is [2, 4]. Defaults to 2.

        ## Documentation reference:
            https://learn.microsoft.com/en-gb/minecraft/creator/reference/content/blockreference/examples/blocktraits
        """
        if direction not in ("up", "down"):
            raise ValueError("multi_block direction must be either 'up' or 'down'.")
        if not 2 <= part_count <= 4:
            raise ValueError("multi_block part_count must be between 2 and 4.")

        self._traits["minecraft:multi_block"] = {
            "enabled_states": ["minecraft:multi_block_part"],
            "parts": part_count,
            "direction": direction,
        }

        return self

    @property
    def export(self):
        return self._traits


class _BlockServerDescription(MinecraftDescription):
    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """The block description.

        Parameters:
            name (str): The name of the block.
            is_vanilla (bool, optional): Whether or not the block is a vanilla block. Defaults to False.
        """
        super().__init__(name, is_vanilla)
        self._traits = _BlockTraits()
        self._description["description"].update({"states": {}, "traits": {}, "menu_category": {}})

    def add_state(self, name: str, range: set[float | str | bool]):
        """Adds a state to the block.

        Parameters:
            name (str): The name of the state.
            range (float | str |bool): Values this state can have.

        """
        if len(range) > 16:
            raise ValueError(f"A block state can only have up to 16 values. {self._object_type}[{self.name}].")

        state_name = name if ":" in name else f"{CONFIG.NAMESPACE}:{name}"
        self._description["description"]["states"][state_name] = range
        return self

    def menu_category(
        self,
        category: ItemCategory = ItemCategory.none,
        group: ItemGroups | str = ItemGroups.none,
        is_hidden_in_commands: bool = False,
    ):
        """Sets the menu category for the Block.

        Parameters:
            category (ItemCategory, optional): The category of the Block. Defaults to ItemCategory.none.
            group (str, optional): The group of the Block. Defaults to None.
            is_hidden_in_commands (bool, optional): Whether the Block is hidden in commands. Defaults to False.

        ## Documentation reference:
            https://learn.microsoft.com/en-gb/minecraft/creator/reference/content/blockreference/examples/blockdescription?view=minecraft-bedrock-stable#menu_category-parameters

        """
        self._description["description"]["menu_category"]["category"] = (
            str(category) if not category == ItemCategory.none else {}
        )
        self._description["description"]["menu_category"]["group"] = str(group) if not group == ItemGroups.none else {}
        self._description["description"]["menu_category"]["is_hidden_in_commands"] = (
            is_hidden_in_commands if is_hidden_in_commands else {}
        )
        return self

    @property
    def traits(self):
        """Sets the traits for the block.

        ## Documentation reference:
            https://learn.microsoft.com/en-gb/minecraft/creator/reference/content/blockreference/examples/blocktraits
        """
        return self._traits

    def _export(self):
        self._description["description"]["traits"] = self._traits.export
        return super()._export()


class _BlockServer(AddonObject):
    """The block server object."""

    _extension = ".block.json"
    _path = os.path.join(CONFIG.BP_PATH, "blocks")
    _object_type = "Block Server"

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """The block server object.

        Parameters:
            name (str): The name of the block.
            is_vanilla (bool, optional): Whether or not the block is a vanilla block. Defaults to False.
        """
        super().__init__(name)
        self._server_block = JsonSchemes.server_block()
        self._description = _BlockServerDescription(name, is_vanilla)
        self._components = _PermutationComponents(None)
        self._permutations: list[_PermutationComponents] = []

    @property
    def description(self):
        """The block description."""
        return self._description

    @property
    def components(self):
        """The block components."""
        return self._components

    def permutation(self, condition: str | Molang):
        """Adds a permutation to the block.

        Parameters:
            condition (str | Molang): The condition for the permutation.
        """
        self._permutation = _PermutationComponents(condition)
        self._permutations.append(self._permutation)
        return self._permutation

    def _export(self):
        """Queues the block to be exported."""
        self._server_block["minecraft:block"].update(self.description._export())
        self._server_block["minecraft:block"].update(self._components._export())
        comps: dict = self._server_block["minecraft:block"]["components"]
        self._server_block["minecraft:block"]["permutations"] = [
            permutation._export() for permutation in self._permutations
        ]

        if not BlockMaterialInstance._identifier in comps:
            raise RuntimeError(f"Block {self.identifier} missing default component. Block [{self.identifier}]")
        if not BlockGeometry._identifier in comps:
            raise RuntimeError(f"Block {self.identifier} missing at least one geometry. Block [{self.identifier}]")
        if not BlockDisplayName._identifier in self._server_block["minecraft:block"]["components"]:
            self._server_block["minecraft:block"]["components"][BlockDisplayName._identifier] = self._display_name

        self.content(self._server_block)
        super()._export()


class _BlockClient(AddonObject):
    def __init__(self, name: str, is_vanilla: bool = False):
        super().__init__(name, is_vanilla)

    def block_sound(
        self,
        sound_identifier: str,
        sound_event: BlockSoundEvent | BlockInteractiveSoundEvent,
        volume: float = 1.0,
        pitch: tuple[float, float] = (0.8, 1.2),
        max_distance: float = 0,
        min_distance: float = 9999,
    ):
        sound_event_obj = SoundEvent()
        if isinstance(sound_event, BlockSoundEvent):
            sound_def = sound_event_obj.add_block_event(
                self.identifier,
                sound_identifier,
                sound_event,
                volume,
                pitch,
                max_distance,
                min_distance,
            )
            return sound_def
        elif isinstance(sound_event, BlockInteractiveSoundEvent):
            sound_def = sound_event_obj.add_block_interactive_event(
                self.identifier,
                sound_identifier,
                sound_event,
                volume,
                pitch,
                max_distance,
                min_distance,
            )
            return sound_def
        else:
            raise TypeError("Invalid sound event type.")

    def queue(self):
        return


# ===========================================


class Block(MinecraftBlockDescriptor):
    _object_type = "Block"

    def __init__(self, name, is_vanilla=False):
        super().__init__(name, is_vanilla)

        self.server = _BlockServer(name, is_vanilla)
        self.client = _BlockClient(name, is_vanilla)

        self._item = None

    def descriptor(
        self,
        states: Mapping[str, str | int | float | bool] = None,
        tags: list[str] = None,
    ):
        if any([states, tags]):
            return {
                "name": self.identifier,
                "states": states if states else {},
                "tags": tags if tags else {},
            }
        return self.identifier

    @property
    def item(self):
        if not self._item:
            from anvil.api.items.items import Item

            self._item = Item(self.name)

        return self._item

    def queue(self):
        """Queues the block to be exported."""
        self.server.queue()
        self.client.queue()

        if self._item:
            self._item.queue()

        ANVIL._queue(self)

    def _export(self):
        block_name_comp = self.server._server_block["minecraft:block"]["components"][BlockDisplayName._identifier]
        if block_name_comp.startswith("tile."):
            display_name = AnvilTranslator().get_localization_value(block_name_comp)
        else:
            display_name = block_name_comp

        CONFIG.Report.add_report(
            ReportType.BLOCK,
            vanilla=self._is_vanilla,
            col0=display_name,
            col1=self.identifier,
            col2=[
                f"{key}: {[', '.join(str(v) for v in value)]}"
                for key, value in self.server.description._description["description"]["states"].items()
            ],
        )



import os
from typing import Literal

from anvil import CONFIG
from anvil.api.core.enums import Dimension
from anvil.lib.config import ConfigPackageTarget
from anvil.lib.schemas import AddonObject, JsonSchemes


class DimensionConfiguration(AddonObject):
    """A class representing a Fog."""

    _extension = ".json"
    _path = os.path.join(CONFIG.BP_PATH, "dimensions")
    _object_type = "Dimension Height"

    def __init__(self, dimension: Dimension):
        if CONFIG._TARGET == ConfigPackageTarget.ADDON:
            raise RuntimeError("DimensionHeight cannot be used in an addon package.")
        if dimension not in (Dimension.Overworld, Dimension.Nether, Dimension.TheEnd):
            raise ValueError(f"Invalid dimension: {dimension}")

        self.content(JsonSchemes.dimension_configuration(str(dimension)))
        self.generator_type("void")

        super().__init__(str(dimension.removeprefix("minecraft:")))

    def heigh_bounds(self, range: tuple[int, int]) -> "DimensionConfiguration":
        """Set the height bounds for the dimension.

        Args:
            range (tuple[int, int]): A tuple containing the minimum and maximum height.
        Returns:
            DimensionConfiguration: The current DimensionConfiguration instance.
        """
        if len(range) != 2:
            raise ValueError("Range must be a tuple of two integers (min, max).")

        if range[0] >= range[1]:
            raise ValueError("Minimum height must be less than maximum height.")

        if range[0] % 16 != 0 or range[1] % 16 != 0:
            raise ValueError("Height bounds must be multiples of 16.")

        if 512 < range[0] or range[1] < -512:
            raise ValueError("Height bounds must be within -512 to 512.")

        bounds = {"minecraft:dimension_bounds": {"min": range[0], "max": range[1]}}

        self._content["minecraft:dimension"]["components"].update(bounds)

    def generator_type(
        self, generator: Literal["void"] = "void"
    ) -> "DimensionConfiguration":
        """Set the generator type for the dimension. Only 'void' is supported currently.

        Args:
            generator (Literal["void"]): The type of generator to set.
        Returns:
            DimensionConfiguration: The current DimensionConfiguration instance.
        """
        if generator != "void":
            raise ValueError("Currently, only 'void' generator is supported.")

        gen = {"minecraft:generation": {"generator_type": "void"}}

        self._content.update(gen)
        return self



import os
from ast import Dict
from typing import Literal, overload

from anvil import ANVIL
from anvil.api.actors._animations import _BPAnimations
from anvil.api.actors._component_group import _ComponentGroup, _Components, _Properties
from anvil.api.actors._events import _Event
from anvil.api.actors._render_controller import _RenderControllers
from anvil.api.actors.components import EntityInstantDespawn, EntityRideable
from anvil.api.actors.spawn_rules import SpawnRule
from anvil.api.core.core import SoundDefinition, SoundEvent
from anvil.api.core.enums import DamageSensor, Target
from anvil.api.core.sounds import EntitySoundEvent, SoundCategory, _SoundDescription
from anvil.api.core.textures import ItemTexturesObject
from anvil.api.core.types import RGB, RGBA, Vector2D
from anvil.api.logic.molang import Molang, Variable
from anvil.api.pbr.pbr import TextureComponents, TextureSet
from anvil.api.vanilla.entities import MinecraftEntityTypes, vanilla_entity_ids
from anvil.lib.blockbench import BlockBenchSource, _Blockbench
from anvil.lib.config import CONFIG, ConfigPackageTarget
from anvil.lib.lib import MOLANG_PREFIXES
from anvil.lib.reports import ReportType
from anvil.lib.schemas import AddonObject, JsonSchemes, MinecraftDescription, MinecraftEntityDescriptor
from anvil.lib.translator import AnvilTranslator

__all__ = ["Entity", "Attachable"]


class _BP_ControllerState:
    def __init__(self, state_name):
        self._state_name = state_name
        self._controller_state: dict = JsonSchemes.animation_controller_state(self._state_name)
        self._default = True

    def on_entry(self, *commands: str):
        """Events, commands or molang to preform on entry of this state.

        Args:
            *commands (str): The Events, commands or molang to preform on entry of this state.

        Returns:
            This state.
        """
        for command in commands:
            if str(command).startswith(Target.S):
                self._controller_state[self._state_name]["on_entry"].append(command)
            elif any(str(command).startswith(v) for v in MOLANG_PREFIXES):
                self._controller_state[self._state_name]["on_entry"].append(f"{command};")
            else:
                self._controller_state[self._state_name]["on_entry"].append(f"/{command}")
        self._default = False
        return self

    def on_exit(self, *commands: str):
        """Events, commands or molang to preform on exit of this state.

        Args:
            *commands (str): The Events, commands or molang to preform on exit of this state.

        Returns:
            This state.
        """
        for command in commands:
            if str(command).startswith(Target.S.value):
                self._controller_state[self._state_name]["on_exit"].append(f"{command}")
            elif any(str(command).startswith(v) for v in MOLANG_PREFIXES):
                self._controller_state[self._state_name]["on_exit"].append(f"{command};")
            else:
                self._controller_state[self._state_name]["on_exit"].append(f"/{command}")
        self._default = False
        return self

    def animation(self, animation: str, condition: str = None):
        """Animation short name to play during this state.

        Args:
            animation (str): The name of the animation to play.
            condition (str, optional): The condition on which this animation plays.

        Returns:
            This state.
        """
        if condition is None:
            self._controller_state[self._state_name]["animations"].append(animation)
        else:
            self._controller_state[self._state_name]["animations"].append({animation: condition})
        self._default = False
        return self

    def transition(self, state: str, condition: str):
        """Target state to switch to and the condition to do so.

        Parameters
        ----------
        state : str
            The name of the state to transition to.
        condition : str
            The condition that must be met for the transition to occur.

        Returns
        -------
            This state.

        """
        self._controller_state[self._state_name]["transitions"].append({state: str(condition)})
        self._default = False
        return self

    @property
    def _export(self):
        return self._controller_state


class _RP_ControllerState:
    def __init__(self, actor: "_ActorClientDescription", state_name):
        self._actor = actor
        self._state_name = state_name
        self._controller_state = JsonSchemes.animation_controller_state(self._state_name)
        self._controller_state[self._state_name]["particle_effects"] = []
        self._controller_state[self._state_name]["sound_effects"] = []
        self._default = True

    @overload
    def on_entry(self, *commands: str | Molang) -> "_RP_ControllerState": ...

    @overload
    def on_entry(self, variable: Molang | str, value: Molang | str | int | float) -> "_RP_ControllerState": ...

    def on_entry(self, *args, **kwargs) -> "_RP_ControllerState":
        """molang to preform on entry of this state.

        Parameters
        ----------
        *args : str | Molang or (variable, value)
            param args: molang commands to perform on entry of this state, or variable assignment.

        Returns
        -------
            This state.

        """
        # Handle variable assignment case: on_entry(variable, value)
        if len(args) == 2 and not kwargs:
            variable, value = args
            command = f"{variable}={value}"
            if any(command.startswith(v) for v in MOLANG_PREFIXES):
                self._controller_state[self._state_name]["on_entry"].append(f"{command};")
            else:
                raise RuntimeError(
                    f"Invalid variable assignment for on_entry: {command}. Only Molang commands are allowed. Resource Pack Controller State [{self._state_name}]"
                )
        # Handle commands case: on_entry(*commands)
        else:
            for command in args:
                if any(str(command).startswith(v) for v in MOLANG_PREFIXES):
                    self._controller_state[self._state_name]["on_entry"].append(f"{command};")
                else:
                    raise RuntimeError(
                        f"Invalid command for on_entry: {command}. Only Molang commands are allowed. Resource Pack Controller State [{self._state_name}]"
                    )

        self._default = False
        return self

    @overload
    def on_exit(self, *commands: str | Molang) -> "_RP_ControllerState": ...

    @overload
    def on_exit(self, variable: Molang | str, value: Molang | str | int | float) -> "_RP_ControllerState": ...

    def on_exit(self, *args, **kwargs) -> "_RP_ControllerState":
        """molang to preform on entry of this state.

        Parameters
        ----------
        *args : str | Molang or (variable, value)
            param args: molang commands to perform on entry of this state, or variable assignment.

        Returns
        -------
            This state.

        """
        # Handle variable assignment case: on_exit(variable, value)
        if len(args) == 2 and not kwargs:
            variable, value = args
            command = f"{variable}={value}"
            if any(command.startswith(v) for v in MOLANG_PREFIXES):
                self._controller_state[self._state_name]["on_exit"].append(f"{command};")
            else:
                raise RuntimeError(
                    f"Invalid variable assignment for on_exit: {command}. Only Molang commands are allowed. Resource Pack Controller State [{self._state_name}]"
                )
        # Handle commands case: on_exit(*commands)
        else:
            for command in args:
                if any(str(command).startswith(v) for v in MOLANG_PREFIXES):
                    self._controller_state[self._state_name]["on_exit"].append(f"{command};")
                else:
                    raise RuntimeError(
                        f"Invalid command for on_exit: {command}. Only Molang commands are allowed. Resource Pack Controller State [{self._state_name}]"
                    )

        self._default = False
        return self

    def animation(self, animation: str, condition: str = None):
        """Animation short name to play during this state.

        Parameters
        ----------
        animation : str
            The name of the animation to play.
        condition : str , optional
            The condition on which this animation plays.

        Returns
        -------
            This state.

        """
        if condition is None:
            self._controller_state[self._state_name]["animations"].append(animation)
        else:
            self._controller_state[self._state_name]["animations"].append({animation: condition})
        self._default = False
        return self

    def transition(self, state: str, condition: str):
        """Target state to switch to and the condition to do so.

        Parameters
        ----------
        state : str
            The name of the state to transition to.
        condition : str
            The condition that must be met for the transition to occur.

        Returns
        -------
            This state.

        """
        self._controller_state[self._state_name]["transitions"].append({state: str(condition)})
        self._default = False
        return self

    def particle(
        self,
        effect: str,
        locator: str = "root",
        pre_anim_script: str = None,
        bind_to_actor: bool = True,
    ):
        """The effect to be emitted during this state.

        Parameters
        ----------
        effect : str
            The shortname of the particle effect to be played, defined in the Client Entity.
        locator : str
            The name of a locator on the actor where the effect should be located.
        pre_anim_script : str , optional
            A molang script that will be run when the particle emitter is initialized.
        bind_to_actor : bool , optional
            Set to false to have the effect spawned in the world without being bound to an actor.

        Returns
        -------
            This state.

        """
        if not self._actor._description["description"]["particle_effects"].get(effect):
            raise RuntimeError(
                f"Particle effect '{effect}' is not defined in the Client Entity for Actor '{self._actor.name}'."
            )
        particle = {"effect": effect, "locator": locator}
        if pre_anim_script is not None:
            particle.update({"pre_effect_script": f"{pre_anim_script};".replace(";;", ";")})
        if bind_to_actor is False:
            particle.update({"bind_to_actor": False})
        self._controller_state[self._state_name]["particle_effects"].append(particle)
        self._default = False
        return self

    def sound_effect(
        self,
        effect: str,
        locator: str = "root",
    ):
        """Collection of sounds to trigger on entry to this animation state.

        Parameters
        ----------
        effect : str
            The shortname of the sound effect to be played, defined in the Client Entity.

        Returns
        -------
            This state.

        """
        self._controller_state[self._state_name]["sound_effects"].append({"effect": effect, "locator": locator})
        self._default = False
        return self

    def blend_transition(self, blend_value: float):
        """Sets the amount of time to fade out if the animation is interrupted.

        Parameters
        ----------
        blend_value : float
            Blend out time.

        Returns
        -------
            This state.

        """
        self._controller_state[self._state_name]["blend_transition"] = blend_value
        self._default = False
        return self

    @property
    def blend_via_shortest_path(self):
        """When blending a transition to another state, animate each euler axis through the shortest rotation, instead of by value.

        Returns
        -------
            This state.

        """
        self._controller_state[self._state_name]["blend_via_shortest_path"] = True
        self._default = False
        return self

    @property
    def _export(self):
        return self._controller_state


class _BP_Controller:
    def __init__(self, identifier, controller_shortname):
        self._identifier = identifier
        self._controller_shortname = controller_shortname
        self._controllers = JsonSchemes.animation_controller(self._identifier, self._controller_shortname)
        self._controller_states: list[_BP_ControllerState] = []
        self._controller_namespace = (
            f"controller.animation.{self._identifier.replace(':', '.')}.{self._controller_shortname}"
        )
        self._states_names = []
        self._side = "Server"

    def add_state(self, state_name: str):
        self._states_names.append(state_name)
        """Adds a new state to the animation controller.
        
        Parameters
        ----------
        state_name : str
            The name of the state to add.
        
        Returns
        -------
            Animation controller state.
        
        """
        self._controller_state = _BP_ControllerState(state_name)
        self._controller_states.append(self._controller_state)
        return self._controller_state

    @property
    def _export(self):
        collected_states = []
        for state in self._controller_states:
            if not state._default:
                self._controllers[self._controller_namespace]["states"].update(state._export)
                for tr in state._export.values():
                    if "transitions" in tr:
                        for st in tr["transitions"]:
                            collected_states.extend(st.keys())

        for state in set(collected_states):
            if state not in self._states_names:
                raise RuntimeError(
                    f"State '{state}' is referenced in transitions but not defined in the. behavior Pack Animation Controller[{self._identifier}]."
                )

        if len(self._controllers[self._controller_namespace]["states"].items()) > 0:
            return self._controllers
        return {}


class _RP_Controller(_BP_Controller):
    def __init__(self, actor: "_ActorClientDescription", name, controller_shortname):
        super().__init__(name, controller_shortname)
        self._actor = actor
        self._side = "Client"
        self._controller_states: list[_RP_ControllerState] = []

    def add_state(self, state_name: str):
        self._states_names.append(state_name)
        self._controller_state = _RP_ControllerState(self._actor, state_name)
        self._controller_states.append(self._controller_state)
        return self._controller_state


class _BP_AnimationControllers(AddonObject):
    _extension = ".bp_ac.json"
    _path = os.path.join(
        CONFIG.BP_PATH,
        "animation_controllers",
    )
    _object_type = "Behavior Pack Animation Controller"

    def __init__(self, identifier) -> None:
        super().__init__(identifier)
        self._animation_controllers = JsonSchemes.animation_controllers()
        self._controllers_list: list[_BP_Controller] = []

    def add_controller(self, controller_shortname: str) -> _BP_Controller:
        """Adds a new animation controller to the current actor with `default` as the `initial_state`.

        Parameters
        ----------
        controller_shortname : str
            The shortname of the controller you want to add.

        Returns
        -------
            Animation controller.

        """
        ctrl = _BP_Controller(self.identifier, controller_shortname)
        self._controllers_list.append(ctrl)
        return ctrl

    def queue(self, directory: str = None):
        if len(self._controllers_list) > 0:
            for controller in self._controllers_list:
                self._animation_controllers["animation_controllers"].update(controller._export)
            self.content(self._animation_controllers)
            return super().queue(directory=directory)


class _RP_AnimationControllers(AddonObject):
    _extension = ".rp_ac.json"
    _path = os.path.join(CONFIG.RP_PATH, "animation_controllers")
    _object_type = "Resource Pack Animation Controller"

    def __init__(self, actor: "_ActorClientDescription", name) -> None:
        super().__init__(name)
        self._actor = actor
        self._animation_controllers = JsonSchemes.animation_controllers()
        self._controllers_list: list[_RP_Controller] = []

    def add_controller(self, controller_shortname: str) -> _RP_Controller:
        """Adds a new animation controller to the current actor with `default` as the `initial_state`.

        Parameters
        ----------
        controller_shortname : str
            The shortname of the controller you want to add.

        Returns
        -------
            Animation controller.

        """
        self._animation_controller = _RP_Controller(self._actor, self.identifier, controller_shortname)
        self._controllers_list.append(self._animation_controller)
        return self._animation_controller

    def _is_populated(self):
        return len(self._controllers_list) > 0 and any(
            [len(t._controller_states) > 0 for t in [c for c in self._controllers_list]]
        )

    def queue(self, directory: str = None):
        if self._is_populated():
            for controller in self._controllers_list:
                self._animation_controllers["animation_controllers"].update(controller._export)
            self.content(self._animation_controllers)
            return super().queue(directory=directory)


class _ActorReuseAssets:
    def __init__(self, client: "_ActorClientDescription") -> None:
        self.client = client

    def animation(
        self,
        shortname: str,
        animation_name: str,
        animate: bool = False,
        condition: str | Molang = None,
    ):
        if animate is True:
            if condition is None:
                self.client._animate_append(shortname)
            else:
                self.client._animate_append({shortname: condition})

        self.client._description["description"]["animations"].update({shortname: animation_name})

    def animation_controller(
        self,
        shortname: str,
        animation_controller_name: str,
        animate: bool = False,
        condition: str | Molang = None,
    ):
        if animate is True:
            if condition is None:
                self.client._animate_append(shortname)
            else:
                self.client._animate_append({shortname: condition})

        self.client._description["description"]["animations"].update({shortname: animation_controller_name})

    def texture(self, shortname: str, texture_name: str):
        self.client._description["description"]["textures"].update({shortname: texture_name})

    def geometry(self, shortname: str, geometry_name: str):
        self.client._description["description"]["geometry"].update({shortname: geometry_name})

    def particle_effect(self, shortname: str, particle_name: str):
        self.client["description"]["particle_effects"].update({shortname: particle_name})

    def sound_effect(self, shortname: str, sound_name: str):
        self.client._description["description"]["sound_effects"].update({shortname: sound_name})

    def spawn_egg(self, item_sprite: str, texture_index: int = 0):
        self.client._description["description"]["spawn_egg"] = {
            "texture": item_sprite,
            "texture_index": texture_index,
        }

    def render_controller(self, controller_name: str, condition: str = None):
        if condition is None:
            self.client._description["description"]["render_controllers"].append(controller_name)
        else:
            self.client._description["description"]["render_controllers"].append({controller_name: condition})


class _ActorDescription(MinecraftDescription):
    """Base class for all actor descriptions."""

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """Base class for all actor descriptions.

        Parameters:
            name (str): The name of the actor.
            is_vanilla (bool, optional): Whether or not the actor is a vanilla actor. Defaults to False.
        """
        super().__init__(name, is_vanilla)
        self._animation_controllers: _RP_AnimationControllers
        self._description["description"].update({"animations": {}, "scripts": {"animate": []}})

    def _animate_append(self, key: str | dict):
        """Appends a key to the animate list."""

        if key not in self._description["description"]["scripts"]["animate"]:
            self._description["description"]["scripts"]["animate"].append(key)

    def _animation_controller(self, controller_shortname: str, animate: bool = False, condition: str = None):
        """Sets the mapping of internal animation controller references to actual animations.

        Parameters
        ----------
        controller_shortname : str
            The name of the animation controller.
        animate : bool, optional
            bool = False
        condition : str, optional
            str = None

        """

        if animate is True:
            if condition is None:
                self._animate_append(controller_shortname)
            else:
                self._animate_append({controller_shortname: condition})

        self._description["description"]["animations"].update(
            {controller_shortname: f"controller.animation.{CONFIG.NAMESPACE}.{self.name}.{controller_shortname}"}
        )

    def _animations(
        self,
        geometry_name: str,
        animation_shortname: str,
        animate: bool = False,
        condition: str = None,
    ):
        """Sets the mapping of internal animation references to actual animations.

        Parameters
        ----------
        animation_shortname : str
            The name of the animation.
        animate : bool, optional
            bool = False
        condition : str, optional
            str = None

        """

        if animate is True:
            if condition is None:
                self._animate_append(animation_shortname)
            else:
                self._animate_append({animation_shortname: condition})

        self._description["description"]["animations"].update(
            {animation_shortname: f"animation.{CONFIG.NAMESPACE}.{geometry_name}.{animation_shortname}"}
        )


class _ActorClientDescription(_ActorDescription):
    _queued_animation_controllers = set()
    _type = "entity"

    def _render_append(self, key):
        if key not in self._description["description"]["render_controllers"]:
            self._description["description"]["render_controllers"].append(key)

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """Base class for all client actor descriptions.

        Parameters:
            name (str): The name of the actor.
            is_vanilla (bool, optional): Whether or not the actor is a vanilla actor. Defaults to False.
            type (str, optional): The type of the actor. Defaults to "entity".
        """
        super().__init__(name, is_vanilla)
        if self._type not in ["entity", "attachable"]:
            raise RuntimeError(
                f"Invalid type '{self._type}' for actor description. Expected 'entity' or 'attachables'. Actor [{self.identifier}]"
            )

        if is_vanilla and self.identifier not in vanilla_entity_ids:
            raise RuntimeError(
                f"Invalid vanilla entity '{self.identifier}'. Please use a valid vanilla entity from MinecraftEntityTypes. Actor [{self.identifier}]"
            )

        self._is_dummy = False
        self._animation_controllers = _RP_AnimationControllers(self, self._name)
        self._render_controllers = _RenderControllers(self._name, self._type)
        self._texture_set: TextureSet = None

        self._description["description"].update(JsonSchemes.client_description())

    def animation_controller(
        self,
        controller_shortname: str,
        animate: bool = False,
        condition: str | Molang = None,
    ):
        """Sets the mapping of internal animation controller references to actual animations.

        Parameters:
            controller_shortname (str): The name of the animation controller.
            animate (bool, optional): Whether or not to animate the animation controller. Defaults to False.
            condition (str| Molang, optional): The condition to animate the animation controller. Defaults to None.

        """
        self._animation_controller(controller_shortname, animate, condition)
        return self._animation_controllers.add_controller(controller_shortname)

    def animation(
        self,
        blockbench_name: str,
        animation_name: str,
        animate: bool = False,
        condition: str | Molang = None,
    ):
        """Sets the mapping of internal animation references to actual animations.

        Parameters:
            animation_name (str): The name of the animation.
            animate (bool, optional): Whether or not to animate the animation. Defaults to False.
            condition (str | Molang, optional): The condition to animate the animation. Defaults to None.

        """
        bb = _Blockbench(blockbench_name, BlockBenchSource.ACTOR)
        bb.animations.queue_animation(animation_name)

        self._animations(blockbench_name, animation_name, animate, condition)

        return self

    def material(self, material_id: str, material_name: str):
        """This method manages the materials for an entity.

        Parameters:
            material_id (str): The id of the material.
            material_name (str): The name of the material.

        """
        self._description["description"]["materials"].update({str(material_id): str(material_name)})
        return self

    def geometry(self, geometry_name: str, override_bounding_box: Vector2D = None):
        """
        This method manages the geometry for an entity.

        Parameters:
            geometry_name (str): The name of the geometry.
            override_bounding_box (Vector2D, optional): The bounding box to override the default bounding box. Defaults to None.

        Returns:
            self: Returns an instance of the class.
        """

        bb = _Blockbench(geometry_name, BlockBenchSource.ACTOR)
        if override_bounding_box:
            bb.override_bounding_box(override_bounding_box)
        bb.model.queue_model()

        self._description["description"]["geometry"].update(
            {geometry_name: f"geometry.{CONFIG.NAMESPACE}.{geometry_name}"}
        )

        return self

    def dummy(self):
        """This method manages the dummy for an entity."""
        self._is_dummy = True
        return self

    def queryable_geometry(self, geometry_shortname: str):
        """This method manages the queryable geometry for an entity.

        Parameters:
            geometry_shortname (str): The shortname of the geometry.

        """
        if not geometry_shortname in self._description["description"]["geometry"]:
            raise RuntimeError(
                f"Queryable geometry '{geometry_shortname}' not found in entity {self.identifier}. Entity [{self.identifier}]"
            )
        self._description["description"]["queryable_geometry"] = geometry_shortname
        return self

    def texture(
        self,
        blockbench_name: str,
        component: TextureComponents,
    ):
        """This method manages the textures for an entity.

        Parameters:
            blockbench_name (str): The name of the texture.
            texture_name (str): The name of the texture.
        """
        bb = _Blockbench(blockbench_name, BlockBenchSource.ACTOR)
        bb.textures.queue_texture(component.color)

        self._description["description"]["textures"].update(
            {
                component.color: os.path.join(
                    "textures",
                    CONFIG.NAMESPACE,
                    CONFIG.PROJECT_NAME,
                    "actors",
                    blockbench_name,
                    component.color,
                )
            }
        )

        if component.has_aux():
            self._texture_set = TextureSet(component.color, BlockBenchSource.ACTOR)
            self._texture_set.set_blockbench_textures(blockbench_name, component)
            self._texture_set.queue()
        return self

    def script(self, variable: Variable | Molang, *script: Molang | str):
        """This method manages the scripts for an entity.
        Parameters:
            variable (Variable | str): The variable to set the script to.
            *script (Molang | str): The script to set the variable to.
        """
        chained_scripts = "; ".join(map(str, script))

        self._description["description"]["scripts"]["pre_animation"].append(
            f"{variable}={chained_scripts};".replace(";;", ";")
        )
        return self

    def parent_setup(self, variable: Variable | str, script: Molang | str):
        """This method manages the scripts for an entity."""

        self._description["description"]["scripts"]["parent_setup"].append(
            f"{variable}={str(script)};".replace(";;", ";")
        )
        return self

    def init_vars(self, **vars):
        """Initializes variables for an entity.

        Examples:
            >>> Entity("example").init_vars(x=0, y=5, z=8)
        """
        for k, v in vars.items():
            Variable._set_var(k)
            self._description["description"]["scripts"]["initialize"].append(f"v.{k}={v};")
        return self

    def scale(self, scale: Molang | str = 1):
        """Sets the scale of the entity.

        Parameters:
            scale (str | Molang, optional): The scale of the entity. Defaults to "1".
        """
        if scale != 1:
            self._description["description"]["scripts"]["scale"] = scale

    def should_update_bones_and_effects_offscreen(self, bool: bool = False):
        """Sets whether or not the entity should update bones and effects offscreen.

        Parameters:
            bool (bool, optional): Whether or not the entity should update bones and effects offscreen. Defaults to False.
        """
        self._description["description"]["scripts"]["should_update_bones_and_effects_offscreen"] = str(int(bool))

    def should_update_effects_offscreen(self, bool: bool = False):
        """Sets whether or not the entity should update effects offscreen.

        Parameters:
            bool (bool, optional): Whether or not the entity should update effects offscreen. Defaults to False.
        """
        self._description["description"]["scripts"]["should_update_effects_offscreen"] = str(int(bool))

    def scaleXYZ(self, x: Molang | str = "1", y: Molang | str = "1", z: Molang | str = "1"):
        """Sets the scale of the entity.

        Parameters:
            x (str | Molang, optional): The x scale of the entity. Defaults to "1".
            y (str | Molang, optional): The y scale of the entity. Defaults to "1".
            z (str | Molang, optional): The z scale of the entity. Defaults to "1".
        """
        if not x == "1" or not y == "1" or not z == "1":
            self._description["description"]["scripts"]["scalex"] = str(x)
            self._description["description"]["scripts"]["scaley"] = str(y)
            self._description["description"]["scripts"]["scalez"] = str(z)

    def render_controller(self, controller_name: str, condition: str = None):
        """This method manages the render controllers for an entity.

        Parameters:
            controller_name (str): The name of the render controller.
            condition (str, optional): The condition to render the render controller. Defaults to None.

        """
        if condition is None:
            self._render_append(f"controller.render.{CONFIG.NAMESPACE}.{self.name}.{controller_name}")
        else:
            self._render_append({f"controller.render.{CONFIG.NAMESPACE}.{self.name}.{controller_name}": condition})

        return self._render_controllers.add_controller(controller_name)

    def particle_effect(self, particle: str, component: TextureComponents):
        """This method manages the particle effects for an entity.

        Parameters:
            particle (str): The name of the particle effect.

        """
        from anvil.api.world.particles import Particle

        self._particle_name = particle
        self._description["description"]["particle_effects"].update(
            {self._particle_name: f"{CONFIG.NAMESPACE}:{self._particle_name}"}
        )
        Particle(self._particle_name, component).queue()
        return self

    def sound_effect(
        self,
        sound_identifier: str,
        category: SoundCategory = SoundCategory.Neutral,
        max_distance: int = 0,
        min_distance: int = 9999,
    ):
        """This method manages the sound effects for an entity.

        Parameters:
            sound_shortname (str): The shortname of the sound effect.
            sound_path (str): The path to the sound effects from the assets/sounds folder.
            category (SoundCategory, optional): The category of the sound effect. Defaults to SoundCategory.Neutral.

        """
        self._description["description"]["sound_effects"].update(
            {sound_identifier: f"{CONFIG.NAMESPACE}:{sound_identifier}"}
        )

        sound_definition_object = SoundDefinition()
        return sound_definition_object.sound_reference(
            sound_identifier,
            category,
            max_distance=max_distance,
            min_distance=min_distance,
        )

    def sound_event(
        self,
        sound_reference: str,
        sound_event: EntitySoundEvent,
        category: SoundCategory = SoundCategory.Neutral,
        volume: float = 1.0,
        pitch: tuple[float, float] = (0.8, 1.2),
        max_distance: int = 0,
        min_distance: int = 9999,
        variant_query: Molang = None,
        variant_map: str = None,
    ):
        """This method manages the sound events for an entity.

        Parameters:
            sound_reference (str): The identifier of the sound effect.
            sound_event (EntitySoundEvent): The sound event.
            category (SoundCategory, optional): The category of the sound effect. Defaults to SoundCategory.Neutral.
            volume (float, optional): The volume of the sound effect. Defaults to 1.0.
            pitch (tuple[float, float], optional): The pitch of the sound effect. Defaults to (0.8, 1.2).
            max_distance (int, optional): The maximum distance of the sound effect. Defaults to 0.
            min_distance (int, optional): The minimum distance of the sound effect. Defaults to 9999.
            variant_query (Molang, optional): The variant query of the sound effect. Defaults to None.
            variant_map (str, optional): The variant map of the sound effect. Defaults to None.
        """
        sound_event_obj = SoundEvent()
        return sound_event_obj.add_entity_event(
            self.identifier,
            sound_reference,
            sound_event,
            category,
            volume,
            pitch,
            max_distance,
            min_distance,
            variant_query,
            variant_map,
        )

    def _export(self, directory: str = None):
        """Queues the entity for export.

        Parameters:
            directory (str, optional): The directory to export the entity to. Defaults to None.

        Raises:
            RuntimeError: If the entity does not have a geometry, texture, or render controller.
            Exception: If a geometry is reused but the entity it is reused from has not been queued yet.

        """
        if not self._is_dummy:
            if len(self._description["description"]["geometry"]) == 0:
                raise RuntimeError(
                    f"Entity {self.identifier} missing at least one geometry. Entity [{self.identifier}]"
                )

            if len(self._description["description"]["textures"]) == 0:
                raise RuntimeError(f"Entity {self.identifier} missing at least one texture. Entity [{self.identifier}]")

            if len(self._description["description"]["render_controllers"]) == 0:
                raise RuntimeError(
                    f"Entity {self.identifier} missing at least one render controller. Entity [{self.identifier}]"
                )

            for controller in self._render_controllers._controllers:
                controller._validate(
                    self._description["description"]["textures"].keys(),
                    self._description["description"]["geometry"].keys(),
                    self._description["description"]["materials"].keys(),
                )

            self._render_controllers.queue(directory)
            self._animation_controllers.queue(directory)

        if self._texture_set:
            self._texture_set.queue()

        if self._description["description"]["scripts"]["parent_setup"] != []:
            self._description["description"]["scripts"]["parent_setup"] = ";".join(
                self._description["description"]["scripts"]["parent_setup"]
            )

        for controller in self._animation_controllers._controllers_list:
            if controller._export == {}:
                self._description["description"]["animations"].pop(controller._controller_shortname)
                anims: list[str | dict] = self._description["description"]["scripts"]["animate"]

                if controller._controller_shortname in anims:
                    anims.remove(controller._controller_shortname)
                else:
                    d = next(d for d in anims if isinstance(d, dict) and controller._controller_shortname in d)
                    anims.remove(d)

        return super()._export()


class _EntityClientDescription(_ActorClientDescription):
    """Base class for all client entity descriptions."""

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """Base class for all client entity descriptions.

        Parameters:
            name (str): The name of the entity.
            is_vanilla (bool, optional): Whether or not the entity is a vanilla entity. Defaults to False.
        """
        self._spawn_egg_texture = None
        super().__init__(name, is_vanilla)

    def enable_attachables(self, value: bool = False):
        """This determines if the entity should render attachables such as armor."""
        self._description["description"]["enable_attachables"] = value
        return self

    @property
    def HeldItemIgnoresLighting(self):
        """This determines if the held item should ignore lighting."""
        self._description["description"]["held_item_ignores_lighting"] = True
        return self

    @property
    def HideArmor(self):
        """This determines if the armor should be hidden."""
        self._description["description"]["hide_armor"] = True
        return self

    def spawn_egg(self, texture: TextureComponents, texture_index: int = 0):
        """This method adds a spawn egg texture to the entity.

        Parameters:
            item_sprite (str): The name of the item sprite.
        """
        if texture is None:
            raise ValueError("TextureComponents must be provided")

        ItemTexturesObject().add_item(texture.color, [texture.color])
        texture_set = TextureSet(texture.color, BlockBenchSource.ITEM)
        texture_set.set_item_textures(texture)
        texture_set.queue()

        self._description["description"]["spawn_egg"] = {
            "texture": f"{CONFIG.NAMESPACE}:{texture.color}",
            "texture_index": texture_index if texture_index == 0 else {},
        }

    def spawn_egg_color(self, base_color: str, overlay_color: str):
        """This method adds a spawn egg color to the entity.

        Parameters:
            base_color (str): The base color of the spawn egg.
            overlay_color (str): The overlay color of the spawn egg.
        """
        self._description["description"]["spawn_egg"] = {
            "base_color": base_color,
            "overlay_color": overlay_color,
        }

    def held_item_scale(self, scale: Molang | str = 1):
        """Sets the scale of the held item.

        Parameters:
            scale (str | Molang, optional): The scale of the held item. Defaults to "1".
        """
        if scale != 1:
            self._description["description"]["held_item_scale"] = scale

    def _export(self, directory: str):
        """Queues the entity for export.

        Parameters:
            directory (str): The directory to export the entity to.

        """
        super()._export(directory)
        if "spawn_egg" not in self._description["description"] and not self._is_vanilla:
            self._description["description"]["spawn_egg"] = {
                "base_color": "#FFFFFF",
                "overlay_color": "#000000",
            }

        return self._description


class _AttachableClientDescription(_ActorClientDescription):
    """Base class for all client attachable descriptions."""

    _type = "attachable"

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """Base class for all client attachable descriptions.

        Parameters:
            name (str): The name of the attachable.
            is_vanilla (bool, optional): Whether or not the attachable is a vanilla attachable. Defaults to False.
        """
        super().__init__(name, is_vanilla)

    @property
    def reuse_assets(self):
        """Whether or not the actor should reuse assets from another actor."""
        return _ActorReuseAssets(self._description)


class _EntityServerDescription(_ActorDescription):
    """Base class for all server entity descriptions."""

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """Base class for all server entity descriptions.

        Parameters:
            name (str): The name of the entity.
            is_vanilla (bool, optional): Whether or not the entity is a vanilla entity. Defaults to False.
        """
        super().__init__(name, is_vanilla)
        self._properties = _Properties()
        self._description["description"].update(
            {
                "properties": {},
            }
        )

    @property
    def Summonable(self):
        """Sets whether or not we can summon this entity using commands such as /summon.

        Returns
        -------
            Entity description object.

        """
        self._description["description"]["is_summonable"] = True
        return self

    @property
    def Spawnable(self):
        """Sets whether or not this entity has a spawn egg in the creative ui.

        Returns
        -------
            Entity description object.

        """
        self._description["description"]["is_spawnable"] = True
        return self

    @property
    def Experimental(self):
        """Sets whether or not this entity is experimental. Experimental entities are only enabled when the experimental toggle is enabled.

        Returns
        -------
            Entity description object.

        """
        if CONFIG._EXPERIMENTAL:
            if CONFIG._TARGET == "addon":
                raise RuntimeError(
                    f"Experimental entities are not allowed for packages of type 'addon'. Entity [{self.identifier}]."
                )

        self._description["description"]["is_experimental"] = True
        return self

    def RuntimeIdentifier(self, entity: "MinecraftEntityDescriptor"):
        """Sets the runtime identifier of the entity.

        Parameters:
            entity (MinecraftEntityDescriptor): The vanilla entity to get the runtime identifier from.
        """
        if CONFIG._TARGET != ConfigPackageTarget.ADDON:
            if type(entity) is MinecraftEntityDescriptor:
                if entity._allow_runtime:
                    self._description["description"]["runtime_identifier"] = entity.identifier
                else:
                    raise RuntimeError(
                        f"Entity {entity.identifier} does not allow runtime identifier usage. Entity [{self.identifier}]."
                    )
            else:
                raise TypeError(f"Expected Entity, got {type(entity).__name__}. Entity [{self.identifier}]")
        else:
            raise RuntimeError(
                f"Using runtime is not allowed for packages of type '{CONFIG._TARGET}'. Entity [{self.identifier}]"
            )

    @property
    def add_property(self):
        """Adds a property to the entity."""
        return self._properties

    def spawn_category(self):
        """Sets the spawn category of the entity."""
        self._description["description"]["spawn_category"] = "none"

    @property
    def _export(self):
        """Exports the entity description."""
        self._description["description"]["properties"] = self._properties._export
        return super()._export()


class _EntityServer(AddonObject):
    """Base class for all server entities."""

    _extension = ".behavior.json"
    _path = os.path.join(CONFIG.BP_PATH, "entities")
    _object_type = "Server Entity"

    def _add_despawn_function(self):
        """Adds a despawn function to the entity."""
        self.component_group("despawn").add(EntityInstantDespawn())
        self.event("despawn").add("despawn")

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """Base class for all server entities.

        Parameters:
            name (str): The name of the entity.
            is_vanilla (bool, optional): Whether or not the entity is a vanilla entity. Defaults to False.
        """
        super().__init__(name, is_vanilla)
        self._server_entity = JsonSchemes.entity_server()
        self._description = _EntityServerDescription(self.identifier, self._is_vanilla)
        self._animation_controllers = _BP_AnimationControllers(self.identifier)
        self._animations = _BPAnimations(self.identifier)
        self._spawn_rule = SpawnRule(self.name, self._is_vanilla)
        self._components = _Components()
        self._events: dict[str, _Event] = {}
        self._component_groups: list[_ComponentGroup] = []
        self._vars = []
        self._add_despawn_function()

    @property
    def description(self):
        """Returns the entity description."""
        return self._description

    @property
    def spawn_rule(self):
        """Returns the spawn rule of the entity."""
        return self._spawn_rule

    def animation_controller(
        self,
        controller_shortname: str,
        animate: bool = False,
        condition: str | Molang = None,
    ):
        """Sets the mapping of internal animation controller references to actual animations.

        Parameters:
            controller_shortname (str): The name of the animation controller.
            animate (bool, optional): Whether or not to animate the animation controller. Defaults to False.
            condition (str | Molang, optional): The condition to animate the animation controller. Defaults to None.

        """
        self._description._animation_controller(controller_shortname, animate, condition)
        return self._animation_controllers.add_controller(controller_shortname)

    def animation(
        self,
        animation_name: str,
        loop: bool = False,
        animate: bool = False,
        condition: str | Molang = None,
    ):
        """Sets the mapping of internal animation references to actual animations.

        Parameters:
            animation_name (str): The name of the animation.
            loop (bool, optional): Whether or not the animation should loop. Defaults to False.
            animate (bool, optional): Whether or not to animate the animation. Defaults to False.
            condition (str | Molang, optional): The condition to animate the animation. Defaults to None.

        """
        self._description._animations(self._name, animation_name, animate, condition)
        return self._animations.add_animation(animation_name, loop)

    def init_vars(self, **vars):
        """Initializes variables for an entity."""
        for k, v in vars.items():
            Variable._set_var(k)
            self._vars.extend([f"v.{k}={v}"])

        return self

    def event(
        self,
        event_name: (
            Literal[
                "minecraft:entity_born",
                "minecraft:entity_spawned",
                "minecraft:entity_transformed",
                "minecraft:on_prime",
            ]
            | str
        ),
    ):
        """Adds an event to the entity.
            - `minecraft:entity_born` triggers an event call on an entity when the entity is born via breeding.
            - `minecraft:entity_spawned` triggers an event call on an entity when the entity is spawned in the world.
            - `minecraft:entity_transformed` triggers an event call on an entity when transformed into another entity.
            - `minecraft:on_prime` triggers an event call on an entity when the entity is set to explode.

        Parameters:
            event_name (str): The name of the event.
        """
        if event_name in self._events:
            return self._events[event_name]

        self._events[event_name] = _Event(event_name)
        return self._events[event_name]

    @property
    def components(self):
        """Returns the components of the entity."""
        return self._components

    def component_group(self, component_group_name: str):
        """Adds a component group to the entity.

        Parameters:
            component_group_name (str): The name of the component group.

        """
        self._component_groups.append(_ComponentGroup(component_group_name))
        return self._component_groups[-1]

    def add_basic_components(self):
        """Adds basic server components to the entity.

        This includes:
            - `EntityBreathable`
            - `EntityCollisionBox`
            - `EntityDamageSensor`
            - `EntityHealth`
            - `EntityJumpStatic`
            - `EntityKnockbackResistance`
            - `EntityMovement`
            - `EntityMovementType`
            - `EntityNavigationType`
            - `EntityPhysics`
            - `EntityPushable`
            - `EntityPushThrough`
        """
        from anvil.api.actors.components import (
            EntityBreathable,
            EntityCollisionBox,
            EntityDamageSensor,
            EntityHealth,
            EntityJumpStatic,
            EntityKnockbackResistance,
            EntityMovement,
            EntityMovementType,
            EntityNavigationType,
            EntityPhysics,
            EntityPushable,
            EntityPushThrough,
        )
        from anvil.api.core.enums import DamageCause

        self.components.add(
            EntityJumpStatic(0),
            EntityMovementType.Basic(),
            EntityNavigationType.Walk(),
            EntityMovement(0),
            EntityPhysics(True, True),
            EntityKnockbackResistance(10000),
            EntityHealth(6),
            EntityCollisionBox(1, 1),
            EntityBreathable(),
            EntityDamageSensor().add_trigger(DamageCause.All, DamageSensor.No),
            EntityPushable(False, False),
            EntityPushThrough(1),
        )

    def queue(self, directory: str = None):
        """Queues the entity for export.

        Parameters:
            directory (str, optional): The directory to export the entity to. Defaults to None.
        """
        super().queue(directory=directory)

    def _export(self):
        if len(self._vars) > 0:
            self.animation_controller("variables", True).add_state("default").on_entry(*self._vars)
        self._animations.queue(directory=self._directory)
        self._animation_controllers.queue(directory=self._directory)
        self._spawn_rule.queue(directory=self._directory)

        self._server_entity["minecraft:entity"].update(self.description._export)
        self._server_entity["minecraft:entity"]["components"].update(self._components._export()["components"])

        for event in self._events.values():
            self._server_entity["minecraft:entity"]["events"].update(event._export)

        for component_group in self._component_groups:
            self._server_entity["minecraft:entity"]["component_groups"].update(component_group._export())

        self.content(self._server_entity)

        controllers = list(self._animation_controllers._animation_controllers["animation_controllers"].keys())
        cleared_items = []
        for key, controller in self._description._description["description"]["animations"].items():
            if controller.startswith("controller.") and controller not in controllers:
                cleared_items.append(key)

        for item in cleared_items:
            self._description._description["description"]["animations"].pop(item)

        if self._components._has(EntityRideable) or any(c._has(EntityRideable) for c in self._component_groups):
            AnvilTranslator().add_localization_entry(f"action.hint.exit.{self.identifier}", "Sneak to exit")
        return super()._export()


class _EntityClient(AddonObject):
    """Base class for all client entities."""

    _extension = ".entity.json"
    _path = os.path.join(CONFIG.RP_PATH, "entity")
    _object_type = "Client Entity"

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """Base class for all client entities.

        Parameters:
            name (str): The name of the entity.
            is_vanilla (bool, optional): Whether or not the entity is a vanilla entity. Defaults to False.
        """
        super().__init__(name, is_vanilla)
        self._client_entity = JsonSchemes.entity_client()
        self._description = _EntityClientDescription(self.identifier, self._is_vanilla)

    @property
    def reuse_assets(self):
        """Whether or not the actor should reuse assets from another actor."""
        return _ActorReuseAssets(self._description)

    @property
    def description(self):
        """Returns the entity description."""
        return self._description

    def queue(self, directory: str = None):
        """Queues the entity for export.

        Parameters:
            directory (str, optional): The directory to export the entity to. Defaults to None.
        """
        super().queue(directory=directory)

    def _export(self):
        self._client_entity["minecraft:client_entity"].update(self._description._export(self._directory))

        self.content(self._client_entity)
        return super()._export()


class Entity(MinecraftEntityDescriptor):
    def __init__(self, name, is_vanilla=False, allow_runtime=True):
        super().__init__(name, is_vanilla, allow_runtime)

        self.server = _EntityServer(name, is_vanilla)
        self.client = _EntityClient(name, is_vanilla)

    def queue(
        self,
        directory: str = None,
        display_name: str = None,
        spawn_egg_name: str = None,
    ):
        self._display_name = self._display_name if display_name is None else display_name
        spawn_egg_name = f"Spawn {self._display_name}" if spawn_egg_name is None else spawn_egg_name
        if self._is_vanilla:
            directory = "vanilla"

        AnvilTranslator().add_localization_entry(f"entity.{self.identifier}.name", self._display_name)
        AnvilTranslator().add_localization_entry(f"entity.{self.identifier}<>.name", self._display_name)
        AnvilTranslator().add_localization_entry(
            f"item.spawn_egg.entity.{self.identifier}.name",
            spawn_egg_name,
        )
        self.client.queue(directory)
        self.server.queue(directory)
        ANVIL._queue(self)

    def _export(self):
        CONFIG.Report.add_report(
            ReportType.ENTITY,
            vanilla=self._is_vanilla,
            col0=self._display_name,
            col1=self.identifier,
            col2=[event._event_name for event in self.server._events.values()],
        )


class Attachable(AddonObject):
    _extension = ".attachable.json"
    _path = os.path.join(CONFIG.RP_PATH, "attachables")
    _object_type = "Attachable"

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """Initializes the attachable.

        Parameters:
            name (str): The name of the attachable.
        """
        super().__init__(name, is_vanilla)

        self._attachable = JsonSchemes.attachable()
        self._description = _AttachableClientDescription(self.identifier, False)

    @property
    def description(self):
        """Returns the description of the attachable."""
        return self._description

    def queue(self):
        """Queues the attachable."""

        self._attachable["minecraft:attachable"].update(self._description._export())
        self.content(self._attachable)

        CONFIG.Report.add_report(
            ReportType.ATTACHABLE,
            vanilla=self._is_vanilla,
            col0=self._display_name,
            col1=self.identifier,
        )
        super().queue()

    def __str__(self):
        """Returns the identifier of the attachable."""
        return self.identifier



import json
import os

import amulet
from amulet.api.block import Block as amuletBlock
from amulet.api.block import StringTag
from halo import Halo

from anvil.api.vanilla.blocks import MinecraftBlockTypes
from anvil.lib.config import CONFIG
from anvil.lib.format_versions import MANIFEST_BUILD


# Requires Amulet core to build the world
class LDtk:
    def _load_data(self):
        # Using the builtin json library becasue it's
        # much faster at loading json compared to commentjson
        # and ldtk is a very big json file

        with open(self._file, "r") as File:
            self._data = json.load(File)

        if self._data["defaultGridSize"] != 16:
            raise ValueError("Grid size incompatible, must be 16px.")

    def _collect_tile_data(self):
        # Collecting tile data
        for tileset in self._data["defs"]["tilesets"]:
            self.dictionary["tilesets"][tileset["uid"]] = {}

            for tile in tileset["customData"]:
                self.dictionary["tilesets"][tileset["uid"]][tile["tileId"]] = {
                    "origin": self._default_origin_block.identifier,
                    "properties": "",
                }
                for tile_custom_data in tile["data"].split("\n"):
                    if tile_custom_data.startswith("id="):
                        self.dictionary["tilesets"][tileset["uid"]][tile["tileId"]][
                            "id"
                        ] = tile_custom_data.removeprefix("id=").removesuffix("\n")
                    if tile_custom_data.startswith("origin="):
                        self.dictionary["tilesets"][tileset["uid"]][tile["tileId"]][
                            "origin"
                        ] = tile_custom_data.removeprefix("origin=").removesuffix("\n")
                    if tile_custom_data.startswith("properties="):
                        self.dictionary["tilesets"][tileset["uid"]][tile["tileId"]][
                            "properties"
                        ] = (
                            tile_custom_data.removeprefix("properties=")
                            .removesuffix("\n")
                            .removeprefix("[")
                            .removesuffix("]")
                            .split(",")
                        )

    def _collect_level_data(self):
        # Collecting level Data
        for level in self._data["levels"]:
            self.dictionary["levels"][level["identifier"]] = {
                "origin": [abs(level["worldX"]) // 16, abs(level["worldY"]) // 16],
                "size": [abs(level["pxWid"]) // 16, abs(level["pxHei"]) // 16],
                "layer": level["worldDepth"],
                "entities": [],
                "tiles": [],
            }
            level_data = {}
            # Levels saved in the same file
            if level["layerInstances"]:
                level_data = level["layerInstances"]

            else:
                with open(
                    os.path.join("world", "ldtk", level["externalRelPath"]), "r"
                ) as File:
                    level_data = json.load(File)["layerInstances"]

            # Level Layers
            for layer_ix, layer in enumerate(reversed(level_data)):
                # Divide by 16 to normalize block/entity coordinates to the 16x tile grid
                # Auto Tiles
                # 0: No flips
                # 1: X flip
                # 2: Y flip
                # 3: X and Y flip
                for block in layer["autoLayerTiles"]:
                    self.dictionary["levels"][level["identifier"]]["tiles"].append(
                        {
                            "x": block["px"][0] / 16,
                            "y": block["px"][1] / 16,
                            "id": block["t"],
                            # "rotation": 180 if block["f"] == 3 else 90 if block["f"] == 2 else 270 if block["f"] == 1 else 0,
                            "tileset_id": layer["__tilesetDefUid"],
                        }
                    )
                # Grid Tiles
                for block in layer["gridTiles"]:
                    self.dictionary["levels"][level["identifier"]]["tiles"].append(
                        {
                            "x": block["px"][0] / 16,
                            "y": block["px"][1] / 16,
                            "id": block["t"],
                            "tileset_id": layer["__tilesetDefUid"],
                        }
                    )
                # Entities
                for entity in layer["entityInstances"]:
                    dat = {}
                    # Collecting Entity data per level
                    for tile_custom_data in entity["fieldInstances"]:
                        dat[tile_custom_data["__identifier"]] = tile_custom_data[
                            "__value"
                        ]
                    id = f'{"minecraft" if dat["vanilla"] else CONFIG.NAMESPACE}:{entity["__identifier"].lower()}'
                    self.dictionary["levels"][level["identifier"]]["entities"].append(
                        {
                            "x": entity["px"][0] / 16,
                            "y": entity["px"][1] / 16,
                            "tileset_id": entity["__tile"]["tilesetUid"],
                            "entity_id": id,
                            "data": dat,
                        }
                    )

    def _dump_dictionary(self):
        with open(self._file.split(".")[0] + ".json", "w") as File:
            json.dump(self.dictionary, File)

    def _prepare_level(self):
        world = amulet.load_level(CONFIG._WORLD_PATH)

        # The following deletes EVERYTHING, DO NOT USE IF YOU HAVE HAND BUILT STRUCTURES
        if self._clear_chunks:
            for chunk in world.all_chunk_coords("minecraft:overworld"):
                world.delete_chunk(*chunk, "minecraft:overworld")

        # The following creates chunks that are present only in LDtk
        for level in self.dictionary["levels"].values():
            for x in range(level["origin"][0], level["size"][0], 16):
                for y in range(level["origin"][1], level["size"][1], 16):
                    world.create_chunk(x // 16, y // 16, "minecraft:overworld")

        return world

    def __init__(
        self,
        filename: str,
        *,
        clear_chunks: bool = False,
        default_origin_block=MinecraftBlockTypes.Air(),
    ) -> None:
        """Converts LDtk map to Minecraft worlds.

        Parameters:
            filename (str): filename of the map, no extension.
            clear_chunks (bool, optional): If chunks should be cleared before writing to the map. Defaults to False.
        """
        self._file = os.path.join("world", "ldtk", filename + ".ldtk")
        self.dictionary = {"tilesets": {}, "levels": {}, "entities": {}, "enums": []}
        self._clear_chunks = clear_chunks
        self._default_origin_block = default_origin_block
        self._load_data()
        self._collect_tile_data()
        self._collect_level_data()
        # self._dump_dictionary()

    @Halo(text="Converting LDtk world", spinner="dots")
    def convert(
        self,
        plane: str = "yz",
        offset: tuple[float, float, float] = (0, 0, 0),
        export_entities: bool = True,
    ):
        def map_coordinates(level_origin, x, y, layer):
            if plane in ["yx", "xy"]:
                return (
                    int(level_origin[0] + x + offset[0]),
                    int(-64 + level_origin[1] - y + offset[1]),
                    int(offset[2] - layer),
                )
            elif plane in ["xz", "zx"]:
                return (
                    int(level_origin[0] + x + offset[0]),
                    int(offset[1] - layer),
                    int(level_origin[1] + y + offset[2]),
                )
            else:
                raise ValueError(f"Unsupported plane: {plane}")

        game_version = ("bedrock", [int(i) for i in MANIFEST_BUILD.split(".")])
        world = self._prepare_level()

        for level_name, level in self.dictionary["levels"].items():
            for tile in level["tiles"]:
                functions = {k: [] for k in self.dictionary["enums"]}
                mapped_coords = map_coordinates(
                    level["origin"], tile["x"], tile["y"], level["layer"]
                )

                try:
                    prop_get = (
                        self.dictionary.get("tilesets")
                        .get(tile["tileset_id"])
                        .get(tile["id"])
                        .get("properties")
                    )
                except:
                    raise KeyError(
                        f"Error retrieving custom data for tile id: [{tile['id']}]"
                    )

                properties = {
                    str(k): StringTag(v)
                    for k, v in [p.split("=") for p in prop_get]
                    if type(prop_get) is list
                }

                origin_block = amuletBlock(
                    *self.dictionary["tilesets"][tile["tileset_id"]][tile["id"]][
                        "origin"
                    ].split(":"),
                    properties,
                )
                block = amuletBlock(
                    *self.dictionary["tilesets"][tile["tileset_id"]][tile["id"]][
                        "id"
                    ].split(":"),
                    properties,
                )

                world.set_version_block(
                    *mapped_coords, "minecraft:overworld", game_version, origin_block
                )
                world.set_version_block(
                    *mapped_coords, "minecraft:overworld", game_version, block
                )

        with open(os.path.join("scripts", "javascript", "entities.ts"), "w") as File:
            entities = {}
            if export_entities:
                for level_name, level in self.dictionary["levels"].items():
                    entities[level_name] = []
                    for entity in level["entities"]:
                        mapped_coords = map_coordinates(
                            level["origin"], entity["x"], entity["y"], level["layer"]
                        )

                        point = entity["data"].get("point")
                        if point != None:
                            cy = point.get("cy")
                            cx = point.get("cx")
                            if plane in ["yx", "xy"]:
                                cy = -64 + level["origin"][1] - cy + offset[1]
                            elif plane in ["xz", "zx"]:
                                cy = level["origin"][1] + cy + offset[2]

                            entity["data"]["point"]["cy"] = cy
                            entity["data"]["point"]["cx"] = (
                                cx + level["origin"][0] + offset[0]
                            )

                        entities[level_name].append(
                            {
                                "id": entity["entity_id"],
                                "location": {
                                    "x": mapped_coords[0],
                                    "y": mapped_coords[1],
                                    "z": mapped_coords[2],
                                },
                                "data": entity["data"],
                            }
                        )
            File.write(f"export const entities = {json.dumps(entities, indent=4)}")

        world.save()
        world.close()



import os

from anvil.api.logic.molang import Query
from anvil.lib.config import CONFIG
from anvil.lib.lib import MOLANG_PREFIXES
from anvil.lib.schemas import AddonObject, JsonSchemes


class _BPAnimation:
    def __init__(self, identifier, animation_short_name: str, loop: bool):
        self._identifier = identifier
        self._animation_key = (
            f"animation.{identifier.replace(':', '.')}.{animation_short_name}"
        )
        self._animation_length = 0.05
        self._animation = JsonSchemes.bp_animation(
            self._identifier, animation_short_name, loop
        )

    def timeline(self, timestamp: float, *commands: str):
        """Takes a timestamp and a list of events, command or molang to run at that time.

        Args:
            timestamp (float): The timestamp of the event.
            *commands (str): The Events, commands or molang to run on exit of this state.

        Returns:
            This animation.
        """
        if self._animation_length <= timestamp:
            self._animation_length = timestamp + 0.05

        self._animation[self._animation_key][
            "animation_length"
        ] = self._animation_length
        if timestamp not in self._animation[self._animation_key]["timeline"]:
            self._animation[self._animation_key]["timeline"][timestamp] = []
        for command in commands:
            if str(command).startswith("@s"):
                self._animation[self._animation_key]["timeline"][timestamp].append(
                    f"{command}"
                )
            elif any(str(command).startswith(v) for v in MOLANG_PREFIXES):
                self._animation[self._animation_key]["timeline"][timestamp].append(
                    f"{command};"
                )
            else:
                self._animation[self._animation_key]["timeline"][timestamp].append(
                    f"/{command}"
                )
        return self

    def animation_length(self, animation_length: float):
        """This function sets the length of the animation.

        Args:
            animation_length (float): The length of the animation in seconds.

        Returns:
            This animation.
        """
        self._animation[self._animation_key]["animation_length"] = animation_length
        return self

    def anim_time_update(self, anim_time_update: Query):
        """Sets the animation time update query.

        Args:
            anim_time_update (Query): The animation time update query.

        Returns:
            This animation.
        """
        self._animation[self._animation_key]["anim_time_update"] = anim_time_update
        return self

    @property
    def _export(self):
        """Exports the animation data.

        Returns:
            dict: The animation data.
        """
        return self._animation


class _BPAnimations(AddonObject):
    _extension = ".animation.json"
    _path = os.path.join(CONFIG.BP_PATH, "animations")
    _object_type = "behavior Pack Animation"

    def __init__(self, identifier) -> None:
        super().__init__(identifier)
        self._animations = JsonSchemes.animations_bp()
        self._animations_list: list[_BPAnimation] = []

    def add_animation(self, animation_short_name: str, loop: bool = False):
        """Adds a new animation to the current actor.

        Args:
            animation_short_name (str): The shortname of the animation you want to add.
            loop (bool, optional): If the animation should loop or not. Defaults to False.

        Returns:
            Animation.
        """
        self._animation = _BPAnimation(self.identifier, animation_short_name, loop)
        self._animations_list.append(self._animation)
        return self._animation

    def queue(self, directory: str = None):
        """Queues the animations for export.

        Args:
            directory (str, optional): The directory to export to. Defaults to None.
        """
        if len(self._animations_list) > 0:
            for animation in self._animations_list:
                self._animations["animations"].update(animation._export)
            self.content(self._animations)
            return super().queue(directory=directory)



from enum import StrEnum


class MinecraftEnchantments(StrEnum):
    """An enumeration of all vanilla Minecraft enchantments."""

    AquaAffinity = "minecraft:aqua_affinity"
    BaneOfArthropods = "minecraft:bane_of_arthropods"
    Binding = "minecraft:binding"
    BlastProtection = "minecraft:blast_protection"
    BowInfinity = "minecraft:infinity"
    Breach = "minecraft:breach"
    Channeling = "minecraft:channeling"
    Density = "minecraft:density"
    DepthStrider = "minecraft:depth_strider"
    Efficiency = "minecraft:efficiency"
    FeatherFalling = "minecraft:feather_falling"
    FireAspect = "minecraft:fire_aspect"
    FireProtection = "minecraft:fire_protection"
    Flame = "minecraft:flame"
    Fortune = "minecraft:fortune"
    FrostWalker = "minecraft:frost_walker"
    Impaling = "minecraft:impaling"
    Knockback = "minecraft:knockback"
    Looting = "minecraft:looting"
    Loyalty = "minecraft:loyalty"
    LuckOfTheSea = "minecraft:luck_of_the_sea"
    Lunge = "minecraft:lunge"
    Lure = "minecraft:lure"
    Mending = "minecraft:mending"
    Multishot = "minecraft:multishot"
    Piercing = "minecraft:piercing"
    Power = "minecraft:power"
    ProjectileProtection = "minecraft:projectile_protection"
    Protection = "minecraft:protection"
    Punch = "minecraft:punch"
    QuickCharge = "minecraft:quick_charge"
    Respiration = "minecraft:respiration"
    Riptide = "minecraft:riptide"
    Sharpness = "minecraft:sharpness"
    SilkTouch = "minecraft:silk_touch"
    Smite = "minecraft:smite"
    SoulSpeed = "minecraft:soul_speed"
    SwiftSneak = "minecraft:swift_sneak"
    Thorns = "minecraft:thorns"
    Unbreaking = "minecraft:unbreaking"
    Vanishing = "minecraft:vanishing"
    WindBurst = "minecraft:wind_burst"



from typing import List

from anvil.api.core.components import _BaseComponent
from anvil.api.logic.molang import Molang
from anvil.lib.config import CONFIG


class _Components:
    def __init__(self):
        self._component_group_name = "components"
        self._components: List[_BaseComponent] = []

    def _set(self, component: _BaseComponent):
        self.remove(component)
        self.add(component)

    def _remove(self, component: _BaseComponent):
        if self._has(component):
            self._components.remove(component)

    def _has(self, component: _BaseComponent):
        """Checks if the component is already set."""
        return component._identifier in self._components_list()

    def _components_list(self):
        return [cmp._identifier for cmp in self._components]

    def add(self, *components: _BaseComponent):
        for component in components:
            self._components.append(component)
        return self

    def remove(self, *components: _BaseComponent):
        for component in components:
            self._remove(component)
        return self

    def _export(self):
        component_classes = {component.__class__ for component in self._components}
        cmp_dict = {}
        for component in self._components:
            missing_dependencies = [dep.__name__ for dep in component._dependencies if dep not in component_classes]
            conflicting = [cla.__name__ for cla in component._clashes if cla in component_classes]

            if missing_dependencies:
                raise RuntimeError(
                    f"Component '{component.__class__.__name__}' requires missing dependency "
                    f"{missing_dependencies} in '{self._component_group_name}' group."
                )

            if conflicting:
                raise RuntimeError(
                    f"Component '{component.__class__.__name__}' conflicts with "
                    f"{conflicting} in '{self._component_group_name}' group. Remove the conflicting component(s)."
                )

            cmp_dict.update(component)
        return {self._component_group_name: cmp_dict}


class _ComponentGroup(_Components):
    def __init__(self, component_group_name: str):
        super().__init__()
        self._component_group_name = component_group_name


class _Properties:
    def __init__(self):
        self._properties = {}

    def enum(self, name: str, values: tuple[str], default: str, *, client_sync: bool = False):
        self._properties[f"{CONFIG.NAMESPACE}:{name}"] = {
            "type": "enum",
            "default": default,
            "values": values,
            "client_sync": client_sync,
        }
        return self

    def int(
        self,
        name: str,
        range: tuple[int, int],
        *,
        default: int = 0,
        client_sync: bool = False,
    ):
        self._properties[f"{CONFIG.NAMESPACE}:{name}"] = {
            "type": "int",
            "default": default if isinstance(default, Molang) else int(default),
            "range": range,
            "client_sync": client_sync,
        }
        return self

    def float(
        self,
        name: str,
        range: tuple[float, float],
        *,
        default: float = 0,
        client_sync: bool = False,
    ):
        self._properties[f"{CONFIG.NAMESPACE}:{name}"] = {
            "type": "float",
            "default": float(default),
            "range": [float(f) for f in range],
            "client_sync": client_sync,
        }
        return self

    def bool(self, name: str, *, default: bool = False, client_sync: bool = False):
        self._properties[f"{CONFIG.NAMESPACE}:{name}"] = {
            "type": "bool",
            "default": default,
            "client_sync": client_sync,
        }
        return self

    @property
    def _export(self):
        return self._properties



import os

from anvil.api.core.enums import FogCameraLocation, RenderDistanceType
from anvil.lib.config import CONFIG
from anvil.lib.schemas import AddonObject, JsonSchemes, MinecraftDescription


class _FogDistance:
    """
    Class to handle fog distances based on camera location, color, and other parameters. Allows for setting fog color,
    distance, and transition.

    Attributes:
        _camera_location (FogCameraLocation): Specifies the location of the camera.
        _distance (dict): A dictionary containing the fog properties related to the camera location.
    """

    def __init__(
        self, camera_location: FogCameraLocation = FogCameraLocation.Air
    ) -> None:
        """
        Initialize the _FogDistance instance.

        Parameters:
            camera_location (FogCameraLocation, optional): Enum specifying the location of the camera. Defaults to FogCameraLocation.Air.
        """
        self._camera_location = camera_location.value
        self._distance = {self._camera_location: {}}

    def color(self, color: str):
        """
        Set the color of the fog.

        Parameters:
            color (str | hex): The color to set the fog.

        Returns:
            _FogDistance: Returns self for chaining.
        """
        self._distance[self._camera_location]["fog_color"] = color
        return self

    def distance(
        self,
        fog_start: int,
        fog_end: int,
        render_distance_type: RenderDistanceType = RenderDistanceType.Render,
    ):
        """
        Set the starting and ending distance of the fog along with the render distance type.

        Parameters:
            fog_start (int): The starting distance of the fog.
            fog_end (int): The ending distance of the fog.
            render_distance_type (RenderDistanceType, optional): The type of render distance. Defaults to RenderDistanceType.Render.

        Returns:
            _FogDistance: Returns self for chaining.
        """
        if fog_start >= fog_end:
            raise ValueError(
                f"fog_end: [{fog_end}] must be greater than fog_start: [{fog_start}]."
            )

        self._distance[self._camera_location]["fog_start"] = fog_start
        self._distance[self._camera_location]["fog_end"] = fog_end
        self._distance[self._camera_location][
            "render_distance_type"
        ] = render_distance_type.value
        return self

    def transition_fog(
        self,
        color: str,
        fog_start: int,
        fog_end: int,
        render_distance_type: RenderDistanceType = RenderDistanceType.Render,
    ):
        """
        Set the color, starting and ending distance of the fog for transitioning along with the render distance type.

        Parameters:
            color (str): The color to set the fog.
            fog_start (int): The starting distance of the fog.
            fog_end (int): The ending distance of the fog.
            render_distance_type (RenderDistanceType, optional): The type of render distance. Defaults to RenderDistanceType.Render.

        Returns:
            _FogDistance: Returns self for chaining.
        """
        if fog_start >= fog_end:
            raise ValueError(
                f"fog_end: [{fog_end}] must be greater than fog_start: [{fog_start}]."
            )
        self._distance[self._camera_location]["color"] = color
        self._distance[self._camera_location]["fog_start"] = fog_start
        self._distance[self._camera_location]["fog_end"] = fog_end
        self._distance[self._camera_location][
            "render_distance_type"
        ] = render_distance_type.value
        return self

    def _export(self):
        """
        Return the instance's distance data.

        Returns:
            dict: The distance data of the instance.
        """
        return self._distance


class Fog(AddonObject):
    """A class representing a Fog."""

    _extension = ".fog.json"
    _path = os.path.join(CONFIG.RP_PATH, "fogs")
    _object_type = "Fog"

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """Initializes a Fog instance.

        Parameters:
            name (str): The name of the fog.
            is_vanilla (bool, optional): Whether the fog is a vanilla fog. Defaults to False.
        """
        super().__init__(name)
        self._description = MinecraftDescription(self.identifier, is_vanilla)
        self._fog = JsonSchemes.fog()
        self._locations: list[_FogDistance] = []
        self._volumes = []

    def add_distance_location(
        self, camera_location: FogCameraLocation = FogCameraLocation.Air
    ):
        """Adds a distance location to the fog.

        Parameters:
            camera_location (FogCameraLocation, optional): The camera location of the fog. Defaults to FogCameraLocation.Air.
        """
        self._locations.append(_FogDistance(camera_location))
        return self._locations[-1]

    def add_volume(self):
        pass

    def queue(self):
        """Queues the fog to be exported."""
        for location in self._locations:
            self._fog["minecraft:fog_settings"]["distance"].update(location._export())
        self._fog["minecraft:fog_settings"].update(self._description._export())
        self.content(self._fog)
        return super().queue()



from enum import StrEnum


class MinecraftBiomeTypes(StrEnum):
    BambooJungle = "minecraft:bamboo_jungle"
    BambooJungleHills = "minecraft:bamboo_jungle_hills"
    BasaltDeltas = "minecraft:basalt_deltas"
    Beach = "minecraft:beach"
    BirchForest = "minecraft:birch_forest"
    BirchForestHills = "minecraft:birch_forest_hills"
    BirchForestHillsMutated = "minecraft:birch_forest_hills_mutated"
    BirchForestMutated = "minecraft:birch_forest_mutated"
    CherryGrove = "minecraft:cherry_grove"
    ColdBeach = "minecraft:cold_beach"
    ColdOcean = "minecraft:cold_ocean"
    ColdTaiga = "minecraft:cold_taiga"
    ColdTaigaHills = "minecraft:cold_taiga_hills"
    ColdTaigaMutated = "minecraft:cold_taiga_mutated"
    CrimsonForest = "minecraft:crimson_forest"
    DeepColdOcean = "minecraft:deep_cold_ocean"
    DeepDark = "minecraft:deep_dark"
    DeepFrozenOcean = "minecraft:deep_frozen_ocean"
    DeepLukewarmOcean = "minecraft:deep_lukewarm_ocean"
    DeepOcean = "minecraft:deep_ocean"
    DeepWarmOcean = "minecraft:deep_warm_ocean"
    Desert = "minecraft:desert"
    DesertHills = "minecraft:desert_hills"
    DesertMutated = "minecraft:desert_mutated"
    DripstoneCaves = "minecraft:dripstone_caves"
    ExtremeHills = "minecraft:extreme_hills"
    ExtremeHillsEdge = "minecraft:extreme_hills_edge"
    ExtremeHillsMutated = "minecraft:extreme_hills_mutated"
    ExtremeHillsPlusTrees = "minecraft:extreme_hills_plus_trees"
    ExtremeHillsPlusTreesMutated = "minecraft:extreme_hills_plus_trees_mutated"
    FlowerForest = "minecraft:flower_forest"
    Forest = "minecraft:forest"
    ForestHills = "minecraft:forest_hills"
    FrozenOcean = "minecraft:frozen_ocean"
    FrozenPeaks = "minecraft:frozen_peaks"
    FrozenRiver = "minecraft:frozen_river"
    Grove = "minecraft:grove"
    Hell = "minecraft:hell"
    IceMountains = "minecraft:ice_mountains"
    IcePlains = "minecraft:ice_plains"
    IcePlainsSpikes = "minecraft:ice_plains_spikes"
    JaggedPeaks = "minecraft:jagged_peaks"
    Jungle = "minecraft:jungle"
    JungleEdge = "minecraft:jungle_edge"
    JungleEdgeMutated = "minecraft:jungle_edge_mutated"
    JungleHills = "minecraft:jungle_hills"
    JungleMutated = "minecraft:jungle_mutated"
    LegacyFrozenOcean = "minecraft:legacy_frozen_ocean"
    LukewarmOcean = "minecraft:lukewarm_ocean"
    LushCaves = "minecraft:lush_caves"
    MangroveSwamp = "minecraft:mangrove_swamp"
    Meadow = "minecraft:meadow"
    MegaTaiga = "minecraft:mega_taiga"
    MegaTaigaHills = "minecraft:mega_taiga_hills"
    Mesa = "minecraft:mesa"
    MesaBryce = "minecraft:mesa_bryce"
    MesaPlateau = "minecraft:mesa_plateau"
    MesaPlateauMutated = "minecraft:mesa_plateau_mutated"
    MesaPlateauStone = "minecraft:mesa_plateau_stone"
    MesaPlateauStoneMutated = "minecraft:mesa_plateau_stone_mutated"
    MushroomIsland = "minecraft:mushroom_island"
    MushroomIslandShore = "minecraft:mushroom_island_shore"
    Ocean = "minecraft:ocean"
    PaleGarden = "minecraft:pale_garden"
    Plains = "minecraft:plains"
    RedwoodTaigaHillsMutated = "minecraft:redwood_taiga_hills_mutated"
    RedwoodTaigaMutated = "minecraft:redwood_taiga_mutated"
    River = "minecraft:river"
    RoofedForest = "minecraft:roofed_forest"
    RoofedForestMutated = "minecraft:roofed_forest_mutated"
    Savanna = "minecraft:savanna"
    SavannaMutated = "minecraft:savanna_mutated"
    SavannaPlateau = "minecraft:savanna_plateau"
    SavannaPlateauMutated = "minecraft:savanna_plateau_mutated"
    SnowySlopes = "minecraft:snowy_slopes"
    SoulsandValley = "minecraft:soulsand_valley"
    StoneBeach = "minecraft:stone_beach"
    StonyPeaks = "minecraft:stony_peaks"
    SunflowerPlains = "minecraft:sunflower_plains"
    Swampland = "minecraft:swampland"
    SwamplandMutated = "minecraft:swampland_mutated"
    Taiga = "minecraft:taiga"
    TaigaHills = "minecraft:taiga_hills"
    TaigaMutated = "minecraft:taiga_mutated"
    TheEnd = "minecraft:the_end"
    WarmOcean = "minecraft:warm_ocean"
    WarpedForest = "minecraft:warped_forest"


class MinecraftBiomeTags(StrEnum):
    Animal = "animal"
    BasaltDeltas = "basalt_deltas"
    Beach = "beach"
    BeeHabitat = "bee_habitat"
    Birch = "birch"
    Caves = "caves"
    CherryGrove = "cherry_grove"
    Cold = "cold"
    CrimsonForest = "crimson_forest"
    DarkOak = "dark_oak"
    Deep = "deep"
    DeepDark = "deep_dark"
    Desert = "desert"
    DripstoneCaves = "dripstone_caves"
    Edge = "edge"
    ExtremeHills = "extreme_hills"
    FastFishing = "fast_fishing"
    FlowerForest = "flower_forest"
    Forest = "forest"
    Frozen = "frozen"
    FrozenPeaks = "frozen_peaks"
    Grove = "grove"
    HighSeas = "high_seas"
    Hills = "hills"
    Ice = "ice"
    IcePlains = "ice_plains"
    JaggedPeaks = "jagged_peaks"
    Jungle = "jungle"
    Lakes = "lakes"
    Lukewarm = "lukewarm"
    LushCaves = "lush_caves"
    MangroveSwamp = "mangrove_swamp"
    Mega = "mega"
    Mesa = "mesa"
    Monster = "monster"
    MooshroomIsland = "mooshroom_island"
    Mountain = "mountain"
    Mutated = "mutated"
    Nether = "nether"
    NetherWastes = "nether_wastes"
    NetherwartForest = "netherwart_forest"
    NoLegacyWorldgen = "no_legacy_worldgen"
    Ocean = "ocean"
    Overworld = "overworld"
    OverworldGeneration = "overworld_generation"
    Plains = "plains"
    Plateau = "plateau"
    Rare = "rare"
    River = "river"
    Roofed = "roofed"
    Savanna = "savanna"
    Shore = "shore"
    Slime = "slime"
    SnowySlopes = "snowy_slopes"
    SoulsandValley = "soulsand_valley"
    SpawnEndermen = "spawn_endermen"
    SpawnFewPiglins = "spawn_few_piglins"
    SpawnFewZombifiedPiglins = "spawn_few_zombified_piglins"
    SpawnGhast = "spawn_ghast"
    SpawnMagmaCubes = "spawn_magma_cubes"
    SpawnManyMagmaCubes = "spawn_many_magma_cubes"
    SpawnPiglin = "spawn_piglin"
    SpawnZombifiedPiglin = "spawn_zombified_piglin"
    Stone = "stone"
    SurfaceMineshaft = "surface_mineshaft"
    Swamp = "swamp"
    SwampWaterHugeMushroom = "swamp_water_huge_mushroom"
    Taiga = "taiga"
    TheEnd = "the_end"
    Warm = "warm"
    WarpedForest = "warped_forest"
